/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "opencl/opencl.pmh"

#define FE_OCL_DEBUG			FALSE
#define FE_OCL_PROFILE			(FE_CODEGEN==FE_PROFILE)

#define FE_OCL_DEBUG_VECTORS	16
#define FE_OCL_PRINT_VECTORS	FALSE

//* see anteru.net/blog/2012/11/03/2009/index.html
//* see www.drdobbs.com/parallel/a-gentle-introduction-to-opencl/231002854

namespace fe
{

OpenCLOp::OpenCLOp(void):
	m_started(FALSE)
{
#if FE_OCL_DEBUG
	feLog("OpenCLOp::OpenCLOp\n");
#endif
}

OpenCLOp::~OpenCLOp(void)
{
#if FE_OCL_DEBUG
	feLog("OpenCLOp::~OpenCLOp\n");
#endif

	clReleaseKernel(m_clKernel);
	clReleaseCommandQueue(m_clCommandQueue);
	clReleaseProgram(m_clProgram);
//	clReleaseContext(m_clContext);
}

void OpenCLOp::initialize(void)
{
#if FE_OCL_DEBUG
	feLog("OpenCLOp::initialize\n");
#endif

	catalog<String>("icon")="FE_alpha";

	catalog<String>("domain")="point";
	catalog<String>("domain","label")="Domain";
	catalog<String>("domain","choice:0")="point";
	catalog<String>("domain","label:0")="Points";
	catalog<String>("domain","choice:1")="primitive";
	catalog<String>("domain","label:1")="Primitives";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","clobber")=true;
}

void OpenCLOp::startup(void)
{
	if(m_started)
	{
		return;
	}

#if FE_OCL_PROFILE
	sp<Profiler> spProfiler(new Profiler("OpenCLOp startup"));
	spProfiler->begin();

	sp<Profiler::Profile> spProfileContext(
			new Profiler::Profile(spProfiler,"Get Context"));
	spProfileContext->start();
#endif

	sp<Catalog> spMasterCatalog=registry()->master()->catalog();

	m_deviceId=(cl_device_id)
			spMasterCatalog->catalogOrDefault<void*>("clDeviceId",nullptr);
	if(m_deviceId==nullptr)
	{
		feLog("OpenCLOp::handle no deviceId\n");
		return;
	}

	m_clContext=(cl_context)
			spMasterCatalog->catalogOrDefault<void*>("clContext",nullptr);
	if(m_clContext==nullptr)
	{
		feLog("OpenCLOp::handle no clContext\n");
		return;
	}

	m_clCommandQueue=(cl_command_queue)
			spMasterCatalog->catalogOrDefault<void*>("clCommandQueue",nullptr);
	if(m_clCommandQueue==nullptr)
	{
		feLog("OpenCLOp::handle no clCommandQueue\n");
		return;
	}

	feLog("OpenCLOp::startup clContext %p deviceId %p\n",
			m_clContext,m_deviceId);

#if FE_OCL_DEBUG
	cl_ulong max_alloc=0;
	clGetDeviceInfo(m_deviceId,CL_DEVICE_MAX_MEM_ALLOC_SIZE,
			sizeof(cl_ulong),&max_alloc,nullptr);
	const I32 maxMb=max_alloc>>20;
	feLog("CL_DEVICE_MAX_MEM_ALLOC_SIZE %d Mb\n",maxMb);

	size_t max_device_group_size=0;
	clGetDeviceInfo(m_deviceId,CL_DEVICE_MAX_WORK_GROUP_SIZE,
			sizeof(size_t),&max_device_group_size,nullptr);
	feLog("CL_DEVICE_MAX_WORK_GROUP_SIZE %d\n",max_device_group_size);

	cl_uint max_item_dimensions=0;
	clGetDeviceInfo(m_deviceId,CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,
			sizeof(cl_uint),&max_item_dimensions,nullptr);
	feLog("CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS %d\n",max_item_dimensions);

	size_t max_item_sizes[max_item_dimensions];
	clGetDeviceInfo(m_deviceId,CL_DEVICE_MAX_WORK_ITEM_SIZES,
			sizeof(max_item_sizes),max_item_sizes,nullptr);
	feLog("CL_DEVICE_MAX_WORK_ITEM_SIZES %d %d %d\n",
			max_item_sizes[0],max_item_sizes[1],max_item_sizes[2]);
#endif

#if FE_OCL_PROFILE
	sp<Profiler::Profile> spProfileBuild(
			new Profiler::Profile(spProfiler,"Build Program"));
	spProfileBuild->replace(spProfileContext);
#endif

	const char* source=catalog<String>("source").c_str();
	const size_t sourceLength=strlen(source);
	if(!sourceLength)
	{
		return;
	}

	cl_int clError=0;

	m_clProgram=clCreateProgramWithSource(m_clContext,
			1,&source,&sourceLength,&clError);
#if FE_OCL_DEBUG
	feLog("clCreateProgramWithSource clError %d\n",clError);
#endif

	clError=clBuildProgram(m_clProgram,0,nullptr,nullptr,nullptr,nullptr);
#if FE_OCL_DEBUG
	feLog("clBuildProgram clError %d\n",clError);
#endif

	if(clError<0)
	{
		size_t log_size=0;
		clGetProgramBuildInfo(m_clProgram,m_deviceId,CL_PROGRAM_BUILD_LOG,
				0,nullptr,&log_size);

		char* program_log=new char[log_size+1];
		clGetProgramBuildInfo(m_clProgram,m_deviceId,CL_PROGRAM_BUILD_LOG,
				log_size+1,program_log,nullptr);

		feLog("clGetProgramBuildInfo CL_PROGRAM_BUILD_LOG:\n%s\n",program_log);

		delete[] program_log;
	}

#if FE_OCL_PROFILE
	sp<Profiler::Profile> spProfileKernel(
			new Profiler::Profile(spProfiler,"Create Kernel"));
	spProfileKernel->replace(spProfileBuild);
#endif

	m_clKernel=clCreateKernel(m_clProgram,"deform",&clError);
#if FE_OCL_DEBUG
	feLog("clCreateKernel clError %d\n",clError);
#endif

//~	m_clCommandQueue=clCreateCommandQueue(m_clContext,
//~			m_deviceId,0,&clError);
//~#if FE_OCL_DEBUG
//~	feLog("clCreateCommandQueue %p clError %d\n",m_clCommandQueue,clError);
//~#endif
//~
//~	spMasterCatalog->catalog<void*>("clCommandQueue")=m_clCommandQueue;

#if FE_OCL_PROFILE
	spProfileKernel->finish();
	spProfiler->end();

	feLog("%s:\n%s\n",spProfiler->name().c_str(),
			spProfiler->report().c_str());
#endif

	m_started=TRUE;
}

void OpenCLOp::handle(Record& a_rSignal)
{
	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spPointPosition;
	if(!access(spPointPosition,spOutputAccessible,
			e_point,e_position)) return;
	const I32 pointCount=spPointPosition->count();

	const I32 primitiveCount=
			spOutputAccessible->count(SurfaceAccessibleI::e_primitive);

#if FE_OCL_DEBUG
	feLog("OpenCLOp::handle output %p pointCount %d primitiveCount %d\n",
			sp<SurfaceAccessibleOpenCL>(spOutputAccessible).raw(),
			pointCount,primitiveCount);
#endif

	startup();
	if(!m_started)
	{
		return;
	}

#if FE_OCL_PROFILE
	sp<Profiler> spProfiler(new Profiler("OpenCLOp handle"));
	spProfiler->begin();

	sp<Profiler::Profile> spProfileConvert(
			new Profiler::Profile(spProfiler,"Convert Input"));
	spProfileConvert->start();
#endif

	cl_mem pointMem;
	float* floatArray=nullptr;
	BWORD useTemporaryBuffer=FALSE;

	sp<SurfaceAccessibleOpenCL::Bridge> spPositionBridge;

	sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(spOutputAccessible);
	if(spSurfaceAccessibleOpenCL.isValid())
	{
		spPositionBridge=
				spSurfaceAccessibleOpenCL->lookupBridge("pt:P",TRUE);
		if(spPositionBridge.isValid())
		{
			pointMem=spPositionBridge->getClMem()->m_clMem;
		}
	}

	if(!pointMem)
	{
		floatArray=new float[pointCount*4];
		useTemporaryBuffer=TRUE;

		for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point=
					spPointPosition->spatialVector(pointIndex);

			float* pointData=&floatArray[pointIndex*4];
			pointData[0]=point[0];
			pointData[1]=point[1];
			pointData[2]=point[2];
		}
	}

	cl_int bufferSize=pointCount*4*sizeof(float);
#if FE_OCL_DEBUG
	feLog("  bufferSize %d\n",bufferSize);
#endif

#if FE_OCL_PROFILE
	sp<Profiler::Profile> spProfileBuffer(
			new Profiler::Profile(spProfiler,"Create Buffer"));
	spProfileBuffer->replace(spProfileConvert);
#endif

	cl_int clError=0;

	if(!pointMem)
	{
		pointMem=clCreateBuffer(m_clContext,
				CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,
				bufferSize,floatArray,&clError);

#if FE_OCL_DEBUG
		feLog("  clCreateBuffer clError %d\n",clError);
#endif
	}

#if FE_OCL_DEBUG
	size_t max_kernel_group_size=0;
	clGetKernelWorkGroupInfo(m_clKernel,m_deviceId,
			CL_KERNEL_WORK_GROUP_SIZE,
			sizeof(size_t),&max_kernel_group_size,nullptr);
	feLog("  CL_KERNEL_WORK_GROUP_SIZE %d\n",max_kernel_group_size);
#endif

	I32 argIndex=0;

	String arguments=catalog<String>("arguments");
	String arg=catalog<String>("arguments");
	while(!(arg=arguments.parse()).empty())
	{
		const String source=arg.parse("\"",":");
		if(source=="param")
		{
			const String key=arg.parse("\"",":");

			const std::type_info& typeInfo=catalogTypeInfo(key);

#if FE_OCL_DEBUG
			feLog("  PARAM \"%s\" typeid \"%s\"\n",key.c_str(),typeInfo.name());
#endif

			if(typeInfo==typeid(String))
			{
				//* TODO does OpenCL even have strings?
			}
			else if(typeInfo==typeid(I32))
			{
				const I32 integer=catalog<I32>(key);
				clError=clSetKernelArg(m_clKernel,argIndex++,
						sizeof(integer),&integer);
			}
			else if(typeInfo==typeid(Real))
			{
				const Real real=catalog<Real>(key);
				clError=clSetKernelArg(m_clKernel,argIndex++,
						sizeof(real),&real);
			}
			else if(typeInfo==typeid(SpatialVector))
			{
				const Vector4& rVector4=catalog<SpatialVector>(key);
				clError=clSetKernelArg(m_clKernel,argIndex++,
						sizeof(rVector4),&rVector4);
			}

#if FE_OCL_DEBUG
			feLog("  clSetKernelArg param \"%s\" clError %d\n",
					key.c_str(),clError);
#endif
		}
		else if(source=="output")
		{
			String rate=arg.parse("\"",":");
			String attribute=arg.parse("\"",":");

			if(!rate.empty() && attribute.empty())
			{
				attribute=rate;
				rate="";
			}

#if FE_OCL_DEBUG
			feLog("  OUTPUT \"%s\" \"%s\"\n",rate.c_str(),attribute.c_str());
#endif

			if(rate=="pt" && attribute=="P")
			{
				clError=clSetKernelArg(m_clKernel,argIndex++,
						sizeof(cl_mem),&pointMem);
#if FE_OCL_DEBUG
				feLog("  clSetKernelArg output %p pointMem %p clError %d\n",
						spOutputAccessible.raw(),pointMem,clError);
#endif
				continue;
			}

			if(spSurfaceAccessibleOpenCL.isNull())
			{
				continue;
			}

			const String bridgeName=rate+":"+attribute;

			sp<SurfaceAccessibleOpenCL::Bridge> spAttrBridge=
					spSurfaceAccessibleOpenCL->lookupBridge(bridgeName,FALSE);
			if(spAttrBridge.isNull())
			{
				continue;
			}

			cl_mem attrMem=spAttrBridge->getClMem()->m_clMem;
			if(!attrMem)
			{
				continue;
			}

			clError=clSetKernelArg(m_clKernel,argIndex++,
					sizeof(cl_mem),&attrMem);
#if FE_OCL_DEBUG
			feLog("  clSetKernelArg output %p \"%s\" attrMem %p clError %d\n",
					spOutputAccessible.raw(),bridgeName.c_str(),
					attrMem,clError);
#endif
		}
	}

#if FE_OCL_DEBUG_VECTORS>0
	const I32 debugSize=FE_OCL_DEBUG_VECTORS*sizeof(Vector4);
	const I32 debugCount=FE_OCL_DEBUG_VECTORS*4;
	float debugArray[debugCount];
	for(I32 m=0;m<debugCount;m++)
	{
		debugArray[m]=0.0;
	}
	cl_mem debugBuffer=clCreateBuffer(m_clContext,
				CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,
				debugSize,debugArray,&clError);

	clError=clSetKernelArg(m_clKernel,argIndex++,
			sizeof(cl_mem),&debugBuffer);
#endif

#if FE_OCL_PROFILE
	sp<Profiler::Profile> spProfileRun(
			new Profiler::Profile(spProfiler,"Run Kernel"));
	spProfileRun->replace(spProfileBuffer);
#endif

	const String domain=catalog<String>("domain");
	const size_t global_work_size=
			(domain=="point")? pointCount: primitiveCount;

	clError=clEnqueueNDRangeKernel(m_clCommandQueue,m_clKernel,1,nullptr,
			&global_work_size,nullptr,0,nullptr,nullptr);
#if FE_OCL_DEBUG
	feLog("  clEnqueueNDRangeKernel clError %d\n",clError);
#endif

#if FE_OCL_DEBUG_VECTORS>0 && FE_OCL_PRINT_VECTORS
	clError=clEnqueueReadBuffer(m_clCommandQueue,debugBuffer,CL_TRUE,0,
			debugSize,debugArray,0,nullptr,nullptr);

	feLog("DEBUG\n");
	for(I32 row=0;row<FE_OCL_DEBUG_VECTORS;row++)
	{
		const I32 row4=row*4;
		const Vector4 vector4(debugArray[row4],debugArray[row4+1],
				debugArray[row4+2],debugArray[row4+3]);

		feLog("  %d: %s\n",row,c_print(vector4));
	}
#endif

#if FE_OCL_PROFILE
	if(useTemporaryBuffer)
	{
		//* NOTE block until kernel is finished (flush doesn't do it)
		clError=clEnqueueReadBuffer(m_clCommandQueue,pointMem,CL_TRUE,0,
				1,floatArray,0,nullptr,nullptr);
	}

	sp<Profiler::Profile> spProfileRead(
			new Profiler::Profile(spProfiler,"Read Back"));
	spProfileRead->replace(spProfileRun);
#endif

	if(useTemporaryBuffer)
	{
		clError=clEnqueueReadBuffer(m_clCommandQueue,pointMem,CL_TRUE,0,
				bufferSize,floatArray,0,nullptr,nullptr);
#if FE_OCL_DEBUG
		feLog("  clEnqueueReadBuffer clError %d\n",clError);
#endif
	}

#if FE_OCL_PROFILE
	sp<Profiler::Profile> spProfileOutput(
			new Profiler::Profile(spProfiler,"Write Output"));
	spProfileOutput->replace(spProfileRead);
#endif

	if(useTemporaryBuffer)
	{
		for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			float* pointData=&floatArray[pointIndex*4];
			const SpatialVector point(pointData[0],pointData[1],pointData[2]);

			spPointPosition->set(pointIndex,point);
		}
	}

#if FE_OCL_PROFILE
	sp<Profiler::Profile> spProfileCleanup(
			new Profiler::Profile(spProfiler,"Cleanup"));
	spProfileCleanup->replace(spProfileOutput);
#endif

	if(useTemporaryBuffer)
	{
		delete[] floatArray;
	}

#if FE_OCL_PROFILE
	spProfileCleanup->finish();
	spProfiler->end();

	feLog("%s:\n%s\n",spProfiler->name().c_str(),
			spProfiler->report().c_str());
#endif

#if FE_OCL_DEBUG
	feLog("OpenCLOp::handle done\n");
#endif
}

} // namespace
