/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define GL_GLEXT_PROTOTYPES

#include "opengl.pmh"

namespace fe
{

#define DGL_VERTEXARRAY		TRUE	// allow render calls in group form
#define DGL_LINEWIDTH		TRUE	// allow custom linewidth
#define DGL_ANTIALIAS_LINE	TRUE	// allow antialiased lines
#define DGL_POINT_ON_LINE	FALSE	// make sure short lines show up
#define DGL_THREAD_LOCKING	TRUE	// lock the event context

#define DGL_CHECK_ERROR		(FE_CODEGEN<=FE_DEBUG)	// check glGetError

		// assert if alpha<=0.0
#define	DGL_CHECK_ALPHA		FALSE //(FE_CODEGEN<=FE_DEBUG)

#if FE_DOUBLE_REAL
#define	DGL_REAL			GL_DOUBLE
#else
#define	DGL_REAL			GL_FLOAT
#endif

FE_DL_PUBLIC sp<DrawOpenGL::Font> DrawOpenGL::ms_spFont;

DrawOpenGL::DrawOpenGL(void):
	m_fontbase(0),
	m_multiplication(0.0),
	m_isDirect(TRUE),
	m_lockThreads(TRUE)
{
	m_spFontHold=ms_spFont;

#if FE_OS==FE_LINUX
	Display *pDisplay=glXGetCurrentDisplay();
	GLXContext glxContext=glXGetCurrentContext();
	m_isDirect=glXIsDirect(pDisplay,glxContext);
	feLog("pDisplay=%p glxContext=%p direct=%d\n",
			pDisplay,glxContext,m_isDirect);
#endif

	if(!m_isDirect)
	{
		feLog("DrawOpenGL: indirect context detected\n");
	}

	fe::String value;
	if(!fe::System::getEnvironmentVariable("FE_GL_LOCKING", value) ||
			!atoi(value.c_str()))
	{
		m_lockThreads=FALSE;
	}

	m_pQuadric=gluNewQuadric();
}

DrawOpenGL::~DrawOpenGL()
{
	gluDeleteQuadric(m_pQuadric);
	m_spFontHold=NULL;

	if(ms_spFont.isValid() && ms_spFont->count()==1)
	{
		ms_spFont=NULL;
	}
}

void DrawOpenGL::initialize(void)
{
	const char* shading=(const char *)glGetString(GL_SHADING_LANGUAGE_VERSION);
	feLog("DrawOpenGL::initialize GLSL version string \"%s\"\n",shading);

	BWORD found_context=FALSE;
#if FE_2DGL==FE_2D_X_GFX	// Linux or X11-mode OSX
	found_context=(glXGetCurrentContext()!=NULL);
#elif FE_OS==FE_WIN32 || FE_OS==FE_WIN64
	found_context=(wglGetCurrentContext()!=NULL);
#endif

	if(!found_context)
	{
		feX(e_cannotCreate,"DrawOpenGL::DrawOpenGL()","no GL context");
	}
	else
	{
		glEnable(GL_DEPTH_TEST);

		createDefaultTexture();
		useDefaultTexture();
		setupLights();
	}
}

DrawOpenGL::Font::Font(Real a_multiplication)
{
	//* NOTE see NativeEventContext::loadFont()

	String name="-*-clean-medium-r-*-*-8-*-*-*-*-*-*-*";

	if(a_multiplication>1)
	{
//		name="-*-terminus-bold-r-*-*-22-*-*-*-*-*-*-*";
		name="-*-lucidatypewriter-bold-r-*-*-20-*-*-*-*-*-*-*";
	}

	Display* pDisplayHandle=XOpenDisplay(NULL);
	m_pFontStruct=XLoadQueryFont(pDisplayHandle,name.c_str());
	m_fontRange=m_pFontStruct->max_char_or_byte2+1;
	m_fontBase=glGenLists(m_fontRange);

	const ::Font fontId=m_pFontStruct->fid;
	const U32 first=m_pFontStruct->min_char_or_byte2;
	const U32 last=m_pFontStruct->max_char_or_byte2;

	glXUseXFont(fontId,first,last-first+1,m_fontBase+first);

	//* NOTE from NativeEventContext::calcFontHeight()
	int direction;			/* direction hint */
	XCharStruct overall;	/* overall char structure */

	//* quick test of sample string "Qy," (should get high and low extents)
	XTextExtents(m_pFontStruct,"Qy,",
			3,&direction,&m_fontAscent,&m_fontDescent,&overall);
}

DrawOpenGL::Font::~Font(void)
{
	if(m_fontBase>=0 && m_fontRange>=0)
	{
		glDeleteLists(m_fontBase,m_fontRange);
	}

	m_fontBase= -1;
	m_fontRange= -1;

	if(m_pFontStruct)
	{
		Display* pDisplayHandle=XOpenDisplay(NULL);
		if(pDisplayHandle)
		{
			XFreeFont(pDisplayHandle,m_pFontStruct);
		}
	}

	m_pFontStruct=NULL;
}

void DrawOpenGL::assureFont(void)
{
	if(!ms_spFont.isValid())
	{
		ms_spFont=new Font(multiplication());
	}

	m_spFontHold=ms_spFont;

	if(ms_spFont.isValid())
	{
		m_fontbase=ms_spFont->m_fontBase;
	}
}

void DrawOpenGL::setFontBase(U32 index)
{
	m_fontbase=index;
}

I32 DrawOpenGL::fontHeight(I32* pAscent,I32* pDescent)
{
	if(!ms_spFont.isValid())
	{
		sp<WindowI> spWindowI=view()->window();
		if(spWindowI.isValid())
		{
			sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
			if(spEventContextI.isValid())
			{
				return spEventContextI->fontHeight(pAscent,pDescent);
			}
		}
		return 0;
	}

	if(pAscent)
	{
		*pAscent=ms_spFont->m_fontAscent;
	}
	if(pDescent)
	{
		*pDescent=ms_spFont->m_fontDescent;
	}

	return ms_spFont->m_fontAscent+ms_spFont->m_fontDescent;
}

I32 DrawOpenGL::pixelWidth(String string)
{
	if(!ms_spFont.isValid() || !ms_spFont->m_pFontStruct)
	{
		sp<WindowI> spWindowI=view()->window();
		if(spWindowI.isValid())
		{
			sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
			if(spEventContextI.isValid())
			{
				return spEventContextI->pixelWidth(string);
			}
		}
		return 0;
	}
	return XTextWidth(ms_spFont->m_pFontStruct,string.c_str(),string.length());
}

Real DrawOpenGL::multiplication(void)
{
	if(m_multiplication<=Real(0))
	{
		const Real dpi=fe::compute_dpi();
		m_multiplication=fe::determine_multiplication(dpi);
	}

	return m_multiplication;
}

void DrawOpenGL::pushMatrixState(void)
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
}

void DrawOpenGL::popMatrixState(void)
{
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glPopAttrib();
}

void DrawOpenGL::setBrightness(Real a_brightness)
{
	DrawCommon::setBrightness(a_brightness);
	setupLights();
}

void DrawOpenGL::setContrast(Real a_contrast)
{
	DrawCommon::setContrast(a_contrast);
	setupLights();
}

void DrawOpenGL::createDefaultView(void)
{
	m_spDefaultView=registry()->create("ViewI.ViewOpenGL");
	m_spDefaultView->setName("DrawOpenGL default "+m_spDefaultView->name());
#if FALSE
	if(!m_spDefaultView.isValid())
	{
		feX(e_cannotCreate,"DrawOpenGL::initialize",
				"failed to create a ViewI.ViewOpenGL");
	}
#endif
}

void DrawOpenGL::threadLock(void)
{
	sp<WindowI> spWindowI=view()->window();
	if(spWindowI.isValid())
	{
		sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
		if(spEventContextI.isValid())
		{
			spEventContextI->threadLock();
		}
	}
}

void DrawOpenGL::threadUnlock(void)
{
	sp<WindowI> spWindowI=view()->window();
	if(spWindowI.isValid())
	{
		sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
		if(spEventContextI.isValid())
		{
			spEventContextI->threadUnlock();
		}
	}
}

void DrawOpenGL::flush(void)
{
	sp<WindowI> spWindowI=view()->window();
	if(spWindowI.isValid())
	{
		spWindowI->makeCurrent();
		spWindowI->swapBuffers();
		spWindowI->clear();
	}
}

void DrawOpenGL::checkForError(String a_location)
{
	GLenum error=glGetError();
	if(error!=GL_NO_ERROR)
	{
		String errorString="<unknown>";

		switch(error)
		{
			case GL_NO_ERROR:
				errorString="GL_NO_ERROR";
				break;
			case GL_INVALID_ENUM:
				errorString="GL_INVALID_ENUM";
				break;
			case GL_INVALID_VALUE:
				errorString="GL_INVALID_VALUE";
				break;
			case GL_INVALID_OPERATION:
				errorString="GL_INVALID_OPERATION";
				break;
			case GL_STACK_OVERFLOW:
				errorString="GL_STACK_OVERFLOW";
				break;
			case GL_STACK_UNDERFLOW:
				errorString="GL_STACK_UNDERFLOW";
				break;
			case GL_OUT_OF_MEMORY:
				errorString="GL_OUT_OF_MEMORY";
				break;
			case GL_TABLE_TOO_LARGE:
				errorString="GL_TABLE_TOO_LARGE";
				break;
		}
		feX("DrawOpenGL::checkForError","OpenGL reported \"%s\" at \"%s\"",
				errorString.c_str(),a_location.c_str());
	}
}

void DrawOpenGL::setDrawMode(sp<DrawMode> spMode)
{
	SAFEGUARD;
	DrawCommon::setDrawMode(spMode);

	const BWORD texture=(drawStyle()==DrawMode::e_solid);

	setup(drawStyle(),isLit(),twoSidedLighting(),texture,
			frontfaceCulling(),backfaceCulling(),zBuffering());
	glPointSize(pointSize());
	bindTexture(textureImage());
}

void DrawOpenGL::pushAntialiasPoints(void)
{
	glPushAttrib(GL_COLOR_BUFFER_BIT |
//			GL_DEPTH_BUFFER_BIT |
			GL_ENABLE_BIT | GL_HINT_BIT | GL_LIGHTING_BIT);

//	glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);

	glEnable(GL_POINT_SMOOTH);
	glShadeModel(GL_FLAT);
//	glDepthMask(GL_FALSE);

//	glEnable(GL_BLEND);
//	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void DrawOpenGL::pushAntialiasLines(void)
{
	glPushAttrib(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT |
			GL_ENABLE_BIT | GL_HINT_BIT | GL_LIGHTING_BIT);

//	glHint(GL_LINE_SMOOTH_HINT, GL_DONT_CARE);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

	glEnable(GL_LINE_SMOOTH);
	glShadeModel(GL_FLAT);
	glDepthMask(GL_FALSE);

	activateBlending();
}

void DrawOpenGL::pushBlending(void)
{
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT | GL_DEPTH_BUFFER_BIT);

//	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
//	glEnable(GL_POLYGON_SMOOTH);

	activateBlending();
}

void DrawOpenGL::activateBlending(void)
{
	glEnable(GL_BLEND);
	if(drawStyle()==DrawMode::e_ghost)
	{
		glDepthMask(GL_FALSE);

		const Real forcedAlpha=0.4;	//* TODO param
		glBlendColor(1.0,1.0,1.0,forcedAlpha);

		//* keep fraction of existing color
//		glBlendFunc(GL_CONSTANT_ALPHA, GL_ONE_MINUS_CONSTANT_ALPHA);

		//* purely cumulative
		glBlendFunc(GL_CONSTANT_ALPHA, GL_ONE);
	}
	else
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
}

void DrawOpenGL::drawPoints(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	BWORD multicolor,const Color *color,
	sp<DrawBufferI> spDrawBuffer)
{
//~	if(!vertices && spDrawBuffer.isValid())
//~	{
//~		sp<Catalog> spCatalog(spDrawBuffer);
//~		if(spCatalog.isValid())
//~		{
//~			vertices=spCatalog->catalog<I32>("cl:pointCount");
//~		}
//~	}

//	FEASSERT(vertex);
	FEASSERT(vertices>0);

//	feLog("DrawOpenGL::drawPoints %p vertices %d buffer %d vbo %d\n",
//			vertex,vertices,
//			spDrawBuffer.isValid(),
//			spDrawBuffer.isValid()? sp<Buffer>(spDrawBuffer)->m_vboVertex: -1);

	const BWORD lit=isLit();

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

	const BWORD normals=normal && lit;
	const BWORD toggle_lighting=!normal && lit;
	if(toggle_lighting)
	{
		glDisable(GL_LIGHTING);
	}

	if(antialias())
	{
		pushAntialiasPoints();
	}

	Real alpha=color? color[0][3]: 1.0f;
	const BWORD blending=(multicolor || alpha!=1.0f);
	if(blending && drawStyle()!=DrawMode::e_foreshadow)
	{
		pushBlending();

		//* NOTE presumptive
		glDepthMask(GL_FALSE);
	}

	const U32 passes=1+(drawStyle()==DrawMode::e_foreshadow);
	for(U32 pass=0;pass<passes;pass++)
	{
		if(pass)
		{
			if(blending)
			{
				pushBlending();
				glDepthMask(GL_FALSE);
			}

			setup(drawStyle(),lit,twoSidedLighting(),FALSE,
					frontfaceCulling(),backfaceCulling(),FALSE);
		}

#if DGL_VERTEXARRAY
//~		if (multicolor && color)
//~		{

//			glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
//			glEnableClientState(GL_VERTEX_ARRAY);
//
//			glColorPointer(4,DGL_REAL,sizeof(Color),color);
//			glEnableClientState(GL_COLOR_ARRAY);
//
//			if(normals)
//			{
//				glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
//				glEnableClientState(GL_NORMAL_ARRAY);
//			}

			beginArrays(vertex,normal,NULL,vertices,
					multicolor,color,spDrawBuffer,normals,blending);

			glDrawArrays(GL_POINTS,0,vertices);

			endArrays(NULL,multicolor,color,spDrawBuffer,normals);

//			if(normals)
//			{
//				glDisableClientState(GL_NORMAL_ARRAY);
//			}
//
//			glDisableClientState(GL_COLOR_ARRAY);
//			glDisableClientState(GL_VERTEX_ARRAY);

//~		}
//~		else
//~		{
//~			if (color)
//~			{
//~#if FE_DOUBLE_REAL
//~				glColor4dv(color[0].temp());
//~#else
//~				//* TODO clean up and change all cases
//~				if(passes>1 && !pass)
//~				{
//~					Color recolor=color[0];
//~					recolor[3]=1.0;
//~					glColor4fv(recolor.temp());
//~				}
//~				else
//~				{
//~					glColor4fv(color[0].temp());
//~				}
//~#endif
//~			}
//~
//~			glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
//~			glEnableClientState(GL_VERTEX_ARRAY);
//~
//~			if(normals)
//~			{
//~				glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
//~				glEnableClientState(GL_NORMAL_ARRAY);
//~			}
//~
//~			glDrawArrays(GL_POINTS,0,vertices);
//~
//~			if(normals)
//~			{
//~				glDisableClientState(GL_NORMAL_ARRAY);
//~			}
//~
//~			glDisableClientState(GL_VERTEX_ARRAY);
//~		}
#else
		glBegin(GL_POINTS);

		if(normals)
		{
			U32 m;
			if(multicolor && color)
			{
				for(m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glColor4dv(color[m].temp());
					glNormal3dv(normal[m].temp());
					glVertex3dv(vertex[m].temp());
#else
					glColor4fv(color[m].temp());
					glNormal3fv(normal[m].temp());
					glVertex3fv(vertex[m].temp());
#endif
				}
			}
			else
			{
				if (color)
				{
#if FE_DOUBLE_REAL
					glColor4dv(color[0].temp());
#else
					glColor4fv(color[0].temp());
#endif
				}
				for (m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
						glNormal3dv(normal[m].temp());
						glVertex3dv(vertex[m].temp());
#else
						glNormal3fv(normal[m].temp());
						glVertex3fv(vertex[m].temp());
#endif
				}
			}
		}
		else
		{
			U32 m;
			if(multicolor && color)
			{
				for(m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glColor4dv(color[m].temp());
					glVertex3dv(vertex[m].temp());
#else
					glColor4fv(color[m].temp());
					glVertex3fv(vertex[m].temp());
#endif
				}
			}
			else
			{
				if (color)
				{
#if FE_DOUBLE_REAL
					glColor4dv(color[0].temp());
#else
					glColor4fv(color[0].temp());
#endif
				}
				for (m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glVertex3dv(vertex[m].temp());
#else
					glVertex3fv(vertex[m].temp());
#endif
				}
			}
		}

		glEnd();
#endif
	}

	if(blending)
	{
		glPopAttrib();
	}

	if(toggle_lighting)
	{
		glEnable(GL_LIGHTING);
	}

	//* restore from point smooth
	if(antialias())
	{
		glPopAttrib();
	}

#if DGL_CHECK_ERROR
	checkForError("DrawOpenGL::drawPoints complete");
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawLines(const SpatialVector* vertex,
	const SpatialVector *normal,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color,
	BWORD multiradius,const Real* radius,
	const Vector3i *element,U32 elementCount,
	sp<DrawBufferI> spDrawBuffer)
{
	const Vector2f* texture=NULL;
	drawLinesInternal(vertex,normal,texture,vertices,strip,
			multicolor,color,multiradius,radius,
			element,elementCount,spDrawBuffer);
}

void DrawOpenGL::drawLinesInternal(const SpatialVector* vertex,
	const SpatialVector* normal,const Vector2f* texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color,
	BWORD multiradius,const Real* radius,
	const Vector3i* element,U32 elementCount,
	sp<DrawBufferI> spDrawBuffer)
{
	if(drawStyle()==DrawMode::e_pointCloud)
	{
		drawPoints(vertex,NULL,vertices,multicolor,color,spDrawBuffer);
		return;
	}

	FEASSERT(vertex);
	FEASSERT(vertices > 1);
	FEASSERT(strip != e_fan);

#if DGL_CHECK_ALPHA
	if(!multicolor && color && (*color)[3]<=0.0f)
	{
		feX("DrawOpenGL::drawLines",
				"alpha<=0 is pointless\n  check your color: %s\n",
				c_print(*color));
	}
#endif

	const BWORD lit=isLit();

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

/*
	feLog("drawLines normal %p lit %d aa %d width %.6G"
			" multicolor %d color %p strip %d lock %d"
			" window %d camera %d element %p elementCount %d\n",
			normal,lit,antialias(),lineWidth(),
			multicolor,color,strip,m_lockThreads,
			view()->window().isValid(),
			view()->cameraDisplay().isValid(),
			element,elementCount);
	feLog("  %s\n",c_print(drawMode()));
*/

	const BWORD normals=normal && lit;
	const BWORD toggle_lighting=!normal && lit;
	if(toggle_lighting)
	{
		glDisable(GL_LIGHTING);
	}

#if DGL_ANTIALIAS_LINE
	if(antialias())
		pushAntialiasLines();
#endif

#if DGL_LINEWIDTH
	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	if(pushLineWidth)
	{
		glPushAttrib(GL_LINE_BIT);
		glLineWidth(lineWidth());
	}
#endif

	Real alpha=color? color[0][3]: 1.0f;
	const BWORD blending=(multicolor || alpha!=1.0f ||
			drawStyle()==DrawMode::e_ghost);
	if(blending && drawStyle()!=DrawMode::e_foreshadow)
	{
		pushBlending();
	}

	if(strip==e_triple)
	{
		setup(DrawMode::e_outline,FALSE,FALSE,FALSE,
				FALSE,FALSE,zBuffering());

		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(-1.0f,-1.0f);
	}

	const U32 passes=1+(drawStyle()==DrawMode::e_foreshadow);
	for(U32 pass=0;pass<passes;pass++)
	{
		if(pass)
		{
			if(blending)
			{
				pushBlending();
			}

			setup(drawStyle(),lit,twoSidedLighting(),FALSE,
					frontfaceCulling(),backfaceCulling(),FALSE);
		}

#if DGL_VERTEXARRAY
//~		if (multicolor && color)
//~		{
//			glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
//			glEnableClientState(GL_VERTEX_ARRAY);
//
//			glColorPointer(4,DGL_REAL,sizeof(Color),color);
//			glEnableClientState(GL_COLOR_ARRAY);
//
//			if(normals)
//			{
//				glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
//				glEnableClientState(GL_NORMAL_ARRAY);
//			}

			beginArrays(vertex,normal,texture,vertices,
					multicolor,color,spDrawBuffer,normals,blending);

			if(strip==e_triple)
			{
#if FALSE
				for(GLint first=0;first<GLint(vertices-2);first+=3)
				{
					glDrawArrays(GL_TRIANGLES,first,3);
				}
#else
				glDrawArrays(GL_TRIANGLES,0,vertices);
#endif
			}
			else if(element && elementCount)
			{
				GLint first[elementCount];
				GLint count[elementCount];

				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					const Vector2i& rElement=element[elementIndex];
					first[elementIndex]=rElement[0];
					count[elementIndex]=rElement[1];
				}

				glMultiDrawArrays(GL_LINE_STRIP,first,count,elementCount);
			}
			else
			{
				glDrawArrays(strip==e_strip? GL_LINE_STRIP: GL_LINES,
						0,vertices);
			}

			endArrays(texture,multicolor,color,spDrawBuffer,normals);

//			if(normals)
//			{
//				glDisableClientState(GL_NORMAL_ARRAY);
//			}
//			glDisableClientState(GL_COLOR_ARRAY);
//			glDisableClientState(GL_VERTEX_ARRAY);
//~		}
//~		else
//~		{
//~			if(color)
//~			{
//~#if FE_DOUBLE_REAL
//~				glColor4dv(color[0].temp());
//~#else
//~				//* TODO clean up and change all cases
//~				if(passes>1 && !pass)
//~				{
//~					Color recolor=color[0];
//~					recolor[3]=1.0;
//~					glColor4fv(recolor.temp());
//~				}
//~				else
//~				{
//~					glColor4fv(color[0].temp());
//~				}
//~#endif
//~			}
//~
//~			glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
//~			glEnableClientState(GL_VERTEX_ARRAY);
//~
//~			if(normals)
//~			{
//~				glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
//~				glEnableClientState(GL_NORMAL_ARRAY);
//~			}
//~
//~			if(strip==e_triple)
//~			{
//~				for(GLint first=0;first<GLint(vertices-2);first+=3)
//~				{
//~					glDrawArrays(GL_TRIANGLES,first,3);
//~				}
//~			}
//~			else
//~			{
//~				glDrawArrays(strip==e_strip? GL_LINE_STRIP: GL_LINES,
//~						0,vertices);
//~			}
//~
//~			if(normals)
//~			{
//~				glDisableClientState(GL_NORMAL_ARRAY);
//~			}
//~			glDisableClientState(GL_VERTEX_ARRAY);
//~		}
#else
		//* TODO e_triple

		glBegin(strip==e_strip? GL_LINE_STRIP: GL_LINES);

		U32 m;

		if(normals)
		{
			if (multicolor && color)
			{
				for (m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glColor4dv(color[m].temp());
					glNormal3dv(normal[m].temp());
					glVertex3dv(vertex[m].temp());
#else
					glColor4fv(color[m].temp());
					glNormal3fv(normal[m].temp());
					glVertex3fv(vertex[m].temp());
#endif
				}
			}
			else
			{
				if (color)
#if FE_DOUBLE_REAL
					glColor4dv(color[0].temp());
#else
					glColor4fv(color[0].temp());
#endif
				for (m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glNormal3dv(normal[m].temp());
					glVertex3dv(vertex[m].temp());
#else
					glNormal3fv(normal[m].temp());
					glVertex3fv(vertex[m].temp());
#endif
				}
			}
		}
		else
		{
			if (multicolor && color)
			{
				for (m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glColor4dv(color[m].temp());
					glVertex3dv(vertex[m].temp());
#else
					glColor4fv(color[m].temp());
					glVertex3fv(vertex[m].temp());
#endif
				}
			}
			else
			{
				if (color)
#if FE_DOUBLE_REAL
					glColor4dv(color[0].temp());
#else
					glColor4fv(color[0].temp());
#endif
				for (m=0; m < vertices; m++)
#if FE_DOUBLE_REAL
					glVertex3dv(vertex[m].temp());
#else
					glVertex3fv(vertex[m].temp());
#endif
			}
		}

		glEnd();
#endif

#if DGL_POINT_ON_LINE
		// make sure it shows up when only a point
#if FE_DOUBLE_REAL
		glColor4dv(color[0].temp());
#else
		glColor4fv(color[0].temp());
#endif
		glBegin(GL_POINTS);
#if FE_DOUBLE_REAL
		glVertex3dv(vertex[0].temp());
#else
		glVertex3fv(vertex[0].temp());
#endif
		glEnd();
#endif
	}

	if(strip==e_triple)
	{
		setup(drawStyle(),lit,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),zBuffering());
		glDisable(GL_POLYGON_OFFSET_LINE);
	}

	if(blending)
	{
		glPopAttrib();
	}

#if DGL_LINEWIDTH
	if(pushLineWidth)
	{
		//* restore line width
		glPopAttrib();
	}
#endif

#if DGL_ANTIALIAS_LINE
	if(antialias())
		glPopAttrib();
#endif

	if(toggle_lighting)
	{
		glEnable(GL_LIGHTING);
	}

#if DGL_CHECK_ERROR
	checkForError("DrawOpenGL::drawLines complete");
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawTriangles(const SpatialVector* vertex,
	const SpatialVector* normal,const Vector2f* texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color,
	const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
	sp<DrawBufferI> spDrawBuffer)
{
	if(drawStyle()==DrawMode::e_wireframe ||
			drawStyle()==DrawMode::e_pointCloud)
	{
		if(strip==e_fan)
		{
			//* not exactly...
			strip=e_strip;
		}
		else
		{
			strip=e_triple;
		}

		drawLinesInternal(vertex,normal,texture,vertices,strip,
				multicolor,color,FALSE,NULL,NULL,0,spDrawBuffer);
	}
	else
	{
		drawPolys(3,vertex,normal,texture,vertices,strip,
				multicolor,color,hullPointMap,hullFacePoint,spDrawBuffer);
	}
}

void DrawOpenGL::drawRectangles(const SpatialVector *vertex,
		U32 vertices,BWORD multicolor,const Color *color)
{
	drawPolys(2,vertex,NULL,NULL,vertices,e_discrete,multicolor,color,
			NULL,NULL,sp<DrawBufferI>(NULL));
}

sp<DrawBufferI> DrawOpenGL::createBuffer(void)
{
	sp<Library> spLibrary(new Library());
	spLibrary->setRegistry(registry());

	sp<DrawOpenGL::Buffer> spDrawBuffer(
			Library::create<DrawOpenGL::Buffer>("DrawOpenGL::Buffer",
			spLibrary.raw()));

	return spDrawBuffer;
}

DrawOpenGL::Buffer::Buffer(void):
	m_vboVertex(0),
	m_vboNormal(0),
	m_vboTexture(0),
	m_vboColor(0)
{
#ifdef FE_GL_OPENCL
	m_clMemVertex=NULL;
#endif
}

DrawOpenGL::Buffer::~Buffer(void)
{
	releaseAll();
}

void DrawOpenGL::Buffer::releaseAll(void)
{
#if DGL_THREAD_LOCKING
	if(!m_vboVertex && !m_vboNormal && !m_vboTexture && !m_vboColor)
	{
		return;
	}

	sp<DrawOpenGL> spDrawOpenGL(m_spDrawI);
	const BWORD lockThreads=
			(spDrawOpenGL.isValid() && spDrawOpenGL->m_lockThreads);

	if(lockThreads)
	{
		spDrawOpenGL->threadLock();
	}
#endif

	if(m_vboVertex)
	{
		glDeleteBuffers(1,&m_vboVertex);
		m_vboVertex=0;
	}
	if(m_vboNormal)
	{
		glDeleteBuffers(1,&m_vboNormal);
		m_vboNormal=0;
	}
	if(m_vboTexture)
	{
		glDeleteBuffers(1,&m_vboTexture);
		m_vboTexture=0;
	}
	if(m_vboColor)
	{
		glDeleteBuffers(1,&m_vboColor);
		m_vboColor=0;
	}

#ifdef FE_GL_OPENCL
	if(m_clMemVertex)
	{
		clReleaseMemObject(m_clMemVertex);
	}
#endif

#if DGL_THREAD_LOCKING
	if(lockThreads)
	{
		spDrawOpenGL->threadUnlock();
	}
#endif
}

void DrawOpenGL::beginArrays(const SpatialVector* vertex,
		const SpatialVector* normal,const Vector2f* texture,U32 vertices,
		BWORD multicolor,const Color* color,sp<DrawBufferI> spDrawBuffer,
		BWORD normals,BWORD blending)
{
	sp<DrawOpenGL::Buffer> spOpenGLBuffer(spDrawBuffer);

//	feLog("DrawOpenGL::beginArrays vertices %d buffer valid %d vbo %p"
//			" UV %d texture %p\n",
//			vertices, spOpenGLBuffer.isValid(),
//			spOpenGLBuffer.isValid()? spOpenGLBuffer->m_vboVertex: 0x0,
//			isUvSpace(),texture);

	const BWORD uvSpace=isUvSpace();
	if(uvSpace && !texture)
	{
		feLog("DrawOpenGL::beginArrays"
				" vertices %d in UV space without UV coordinates\n",
				vertices);
		return;
	}

	if(spOpenGLBuffer.isValid())
	{
		if(spOpenGLBuffer->m_spDrawI!=sp<DrawI>(this))
		{
			spOpenGLBuffer->releaseAll();
			spOpenGLBuffer->m_spDrawI=sp<DrawI>(this);
		}

#ifdef FE_USE_OPENCL
		sp<Catalog> spMasterCatalog=registry()->master()->catalog();
#endif

		if(!spOpenGLBuffer->m_vboVertex)
		{
			glGenBuffers(1,&spOpenGLBuffer->m_vboVertex);
			glBindBuffer(GL_ARRAY_BUFFER,spOpenGLBuffer->m_vboVertex);
			if(uvSpace)
			{
				glBufferData(GL_ARRAY_BUFFER,
						sizeof(Vector2f)*vertices,texture,
						GL_STATIC_DRAW);
			}
			else
			{
				glBufferData(GL_ARRAY_BUFFER,
						sizeof(SpatialVector)*vertices,vertex,
						GL_STATIC_DRAW);
			}

#ifdef FE_GL_OPENCL
			cl_context clContext=
					(cl_context)spMasterCatalog->catalogOrDefault<void*>(
					"clContext",NULL);
			if(clContext)
			{
				cl_int clError=0;
				spOpenGLBuffer->m_clMemVertex=
						clCreateFromGLBuffer(clContext,
						CL_MEM_READ_WRITE,spOpenGLBuffer->m_vboVertex,&clError);
				feLog("clCreateFromGLBuffer clError %d\n",clError);
			}
#endif
		}

#ifdef FE_USE_OPENCL
		if(spOpenGLBuffer->m_vboVertex)
		{
			cl_command_queue clCommandQueue=
					(cl_command_queue)spMasterCatalog->catalogOrDefault<void*>(
					"clCommandQueue2",NULL);

			if(clCommandQueue)
			{
				sp<SurfaceAccessibleOpenCL::ClMem> spClMemPoint=
						spOpenGLBuffer->catalog< sp<Counted> >("cl:pt:P");

				cl_mem clMemPoint=(spClMemPoint.isValid())?
						spClMemPoint->m_clMem: NULL;

				feLog("VBO vertices %d copy catalog point buffer %p\n",
						vertices,clMemPoint);

				if(clMemPoint)
				{
#ifdef FE_GL_OPENCL
					cl_int clError=clEnqueueAcquireGLObjects(
							clCommandQueue,1,&spOpenGLBuffer->m_clMemVertex,
							0,NULL,NULL);
//					feLog("clEnqueueAcquireGLObjects clError %d\n",clError);

					//* convert CL 4-vector to GL 3-vector
					const size_t src_origin[3]={0,0,0};
					const size_t dest_origin[3]={0,0,0};
					const size_t region[3]={12,vertices,1};
					clError=clEnqueueCopyBufferRect(
							clCommandQueue,clMemPoint,
							spOpenGLBuffer->m_clMemVertex,
							src_origin,dest_origin,region,
							16,16*vertices,
							12,12*vertices,
							0,NULL,NULL);
					if(clError)
					{
						feLog("clEnqueueCopyBufferRect clError %d\n",clError);
					}

					clError=clEnqueueReleaseGLObjects(
							clCommandQueue,1,&spOpenGLBuffer->m_clMemVertex,
							0,NULL,NULL);
//					feLog("clEnqueueReleaseGLObjects clError %d\n",clError);

					clError=clFinish(clCommandQueue);
//					feLog("clFinish clError %d\n",clError);
#else
					float* floatArray=new float[vertices*4];
					cl_int clError=clEnqueueReadBuffer(clCommandQueue,
							clMemPoint,CL_TRUE,0,vertices*sizeof(Vector4),
							floatArray,0,NULL,NULL);

					SpatialVector* pReplacement=new SpatialVector[vertices];

					float* pFloat=floatArray;
					for(U32 m=0;m<vertices;m++)
					{
						set(pReplacement[m],pFloat[0],pFloat[1],pFloat[2]);
						pFloat+=4;
					}

					delete[] floatArray;

					glBindBuffer(GL_ARRAY_BUFFER,spOpenGLBuffer->m_vboVertex);
					glBufferData(GL_ARRAY_BUFFER,
							sizeof(SpatialVector)*vertices,pReplacement,
							GL_STATIC_DRAW);

					delete[] pReplacement;
#endif
				}
			}
		}
#endif

		glBindBuffer(GL_ARRAY_BUFFER,spOpenGLBuffer->m_vboVertex);
		glVertexPointer(uvSpace? 2: 3,DGL_REAL,0,NULL);
	}
	else if(uvSpace)
	{
		glVertexPointer(2,GL_FLOAT,sizeof(Vector2f),texture);
	}
	else
	{
		glVertexPointer(3,DGL_REAL,sizeof(SpatialVector),vertex);
	}

	glEnableClientState(GL_VERTEX_ARRAY);

	if(multicolor && color)
	{
		if(spOpenGLBuffer.isValid())
		{
			if(!spOpenGLBuffer->m_vboColor)
			{
//				feLog("new color buffer\n");
				glGenBuffers(1,&spOpenGLBuffer->m_vboColor);
				glBindBuffer(GL_ARRAY_BUFFER,
						spOpenGLBuffer->m_vboColor);
				glBufferData(GL_ARRAY_BUFFER,
						sizeof(Color)*vertices,color,
						GL_STATIC_DRAW);
			}

			glBindBuffer(GL_ARRAY_BUFFER,spOpenGLBuffer->m_vboColor);
			glColorPointer(4,DGL_REAL,0,NULL);
		}
		else
		{
			glColorPointer(4,DGL_REAL,sizeof(Color),color);
		}

		glEnableClientState(GL_COLOR_ARRAY);
	}
	else if(color)
	{
		if(blending)
		{
#if FE_DOUBLE_REAL
			glColor4dv(color[0].temp());
#else
			glColor4fv(color[0].temp());
#endif
		}
		else
		{
#if FE_DOUBLE_REAL
			glColor3dv(color[0].temp());
#else
			glColor3fv(color[0].temp());
#endif
		}
	}

	if(normals)
	{
		if(spOpenGLBuffer.isValid())
		{
			if(!spOpenGLBuffer->m_vboNormal)
			{
//				feLog("new normal buffer\n");
				glGenBuffers(1,&spOpenGLBuffer->m_vboNormal);
				glBindBuffer(GL_ARRAY_BUFFER,
						spOpenGLBuffer->m_vboNormal);
				glBufferData(GL_ARRAY_BUFFER,
						sizeof(SpatialVector)*vertices,normal,
						GL_STATIC_DRAW);
			}

			glBindBuffer(GL_ARRAY_BUFFER,spOpenGLBuffer->m_vboNormal);
			glNormalPointer(DGL_REAL,0,NULL);
		}
		else
		{
			glNormalPointer(DGL_REAL,sizeof(SpatialVector),normal);
		}

		glEnableClientState(GL_NORMAL_ARRAY);
	}

	if(texture)
	{
		if(spOpenGLBuffer.isValid())
		{
			if(!spOpenGLBuffer->m_vboTexture)
			{
//				feLog("new texture buffer\n");
				glGenBuffers(1,&spOpenGLBuffer->m_vboTexture);
				glBindBuffer(GL_ARRAY_BUFFER,
						spOpenGLBuffer->m_vboTexture);
				glBufferData(GL_ARRAY_BUFFER,
						sizeof(Vector2f)*vertices,texture,
						GL_STATIC_DRAW);
			}

			glBindBuffer(GL_ARRAY_BUFFER,spOpenGLBuffer->m_vboTexture);
			glTexCoordPointer(2,DGL_REAL,0,NULL);
		}
		else
		{
			glTexCoordPointer(2,GL_FLOAT,sizeof(Vector2f),texture);
		}

		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	}
}

void DrawOpenGL::endArrays(const Vector2f* texture,
		BWORD multicolor,const Color* color,sp<DrawBufferI> spDrawBuffer,
		BWORD normals)
{
	sp<DrawOpenGL::Buffer> spOpenGLBuffer(spDrawBuffer);

	if(texture)
	{
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	if(normals)
	{
		glDisableClientState(GL_NORMAL_ARRAY);
	}
	if(multicolor && color)
	{
		glDisableClientState(GL_COLOR_ARRAY);
	}
	glDisableClientState(GL_VERTEX_ARRAY);

	if(spOpenGLBuffer.isValid())
	{
		glBindBuffer(GL_ARRAY_BUFFER,0);
	}
}

//* private: not generalized; only well defined in how it is called internally
void DrawOpenGL::drawPolys(U32 grain,const SpatialVector* vertex,
	const SpatialVector* normal,const Vector2f* texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color,
	const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
	sp<DrawBufferI> spDrawBuffer)
{
#if FALSE
	feLog("DrawOpenGL::drawPolys"
			" %d gr %d st %d mc %d v0 <%s> n0 <%s> c0 <%s>\n",
			vertices,grain,(U32)strip,multicolor,
			vertex? print(vertex[0]).c_str(): "NULL",
			normal? print(normal[0]).c_str(): "NULL",
			color? print(color[0]).c_str(): "NULL");
#endif

#if FALSE
	if(hullPointMap && hullFacePoint)
	{
		const I32 pointCount=(*hullPointMap).size();
		const I32 primitiveCount=(*hullFacePoint).size();

		feLog("DrawOpenGL::drawPolys vertices %d hull points %d faces %d\n",
				vertices,pointCount,primitiveCount);

		for(I32 pointIndex=0;pointIndex<pointCount && pointIndex<32;
				pointIndex++)
		{
			feLog("map %d -> %d\n",pointIndex,(*hullPointMap)[pointIndex]);
		}
		for(U32 primitiveIndex=0;primitiveIndex<primitiveCount &&
				primitiveIndex<32;primitiveIndex++)
		{
			feLog("face %d  %s\n",primitiveIndex,
					c_print((*hullFacePoint)[primitiveIndex]));
		}
	}
#endif

	FEASSERT(grain>1);
	FEASSERT(vertices>=grain);
	FEASSERT(strip!=e_discrete || !(vertices%grain));
	FEASSERT(vertex);
//	FEASSERT(drawStyle()==DrawMode::e_outline || normal || grain<3);

#if DGL_CHECK_ALPHA
	if(!multicolor && color && (*color)[3]<=0.0f)
	{
		feX("DrawOpenGL::drawPolys",
				"alpha<=0 is pointless\n  check your color: %s\n",
				c_print(*color));
	}
#endif

#if DGL_CHECK_ERROR
	checkForError("DrawOpenGL::drawPolys start");
#endif

	GLenum primitive=GL_QUADS;
	if(grain==3)
	{
		primitive=(strip==e_discrete)? GL_TRIANGLES:
				((strip==e_strip)? GL_TRIANGLE_STRIP:
				GL_TRIANGLE_FAN);
	}

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	Real alpha=color? color[0][3]: 1.0f;
	const BWORD lit=isLit();
//	const BWORD blending=((multicolor || alpha!=1.0f) && lit);
	const BWORD blending=(multicolor || alpha!=1.0f ||
			drawStyle()==DrawMode::e_ghost);
	if(blending && drawStyle()!=DrawMode::e_foreshadow)
	{
		pushBlending();
	}

	const U32 passes=1+(drawStyle()==DrawMode::e_edgedSolid ||
			drawStyle()==DrawMode::e_foreshadow);

	for(U32 pass=0;pass<passes;pass++)
	{
		if(pass)
		{
#if DGL_LINEWIDTH
			if(pushLineWidth)
			{
				glPushAttrib(GL_LINE_BIT);
				glLineWidth(lineWidth());
			}
#endif

			if(drawStyle()==DrawMode::e_edgedSolid)
			{
				setup(DrawMode::e_outline,FALSE,FALSE,FALSE,
						FALSE,FALSE,zBuffering());

				glEnable(GL_POLYGON_OFFSET_LINE);
				glPolygonOffset(-1.0f, 0.0f);
			}
			else
			{
				FEASSERT(drawStyle()==DrawMode::e_foreshadow);
				if(blending)
				{
					pushBlending();
				}

				setup(drawStyle(),lit,twoSidedLighting(),FALSE,
						frontfaceCulling(),backfaceCulling(),FALSE);
			}
		}

#if DGL_CHECK_ERROR
		checkForError("DrawOpenGL::drawPolys setup");
#endif

#if DGL_VERTEXARRAY
		const BWORD normals=!pass && lit;

		if(grain>2)
		{
			beginArrays(vertex,normal,texture,vertices,
					multicolor,color,spDrawBuffer,normals,blending);

#if DGL_CHECK_ERROR
			checkForError("DrawOpenGL::drawPolys array prep");
#endif

			glDrawArrays(primitive,0,vertices);

			endArrays(texture,multicolor,color,spDrawBuffer,normals);
		}
		else
		{
			if(color)
			{
				if(blending)
				{
#if FE_DOUBLE_REAL
					glColor4dv(color[0].temp());
#else
					glColor4fv(color[0].temp());
#endif
				}
				else
				{
#if FE_DOUBLE_REAL
					glColor3dv(color[0].temp());
#else
					glColor3fv(color[0].temp());
#endif
				}
			}

#if DGL_CHECK_ERROR
			checkForError("DrawOpenGL::drawPolys single color");
#endif

			glBegin(primitive);
			for(U32 m=0; m < vertices; m++)
			{
#if FE_DOUBLE_REAL
				glVertex3d(vertex[m].temp()[0],vertex[m+1].temp()[1],
						vertex[m].temp()[2]);
				glVertex3d(vertex[m].temp()[0],vertex[m].temp()[1],
						vertex[m].temp()[2]);
#else
				glVertex3f(vertex[m].temp()[0],vertex[m+1].temp()[1],
						vertex[m].temp()[2]);
				glVertex3f(vertex[m].temp()[0],vertex[m].temp()[1],
						vertex[m].temp()[2]);
#endif

				m++;
#if FE_DOUBLE_REAL
				glVertex3d(vertex[m].temp()[0],vertex[m-1].temp()[1],
						vertex[m].temp()[2]);
				glVertex3d(vertex[m].temp()[0],vertex[m].temp()[1],
						vertex[m].temp()[2]);
#else
				glVertex3f(vertex[m].temp()[0],vertex[m-1].temp()[1],
						vertex[m].temp()[2]);
				glVertex3f(vertex[m].temp()[0],vertex[m].temp()[1],
						vertex[m].temp()[2]);
#endif
			}
			glEnd();
		}
#else
		glBegin(primitive);

		if(multicolor && color)
		{
			for(U32 m=0; m < vertices; m++)
			{
				if (color[m][3] != 1.0f)
#if FE_DOUBLE_REAL
					glColor4dv(color[m].temp());
#else
					glColor4fv(color[m].temp());
#endif
				else
#if FE_DOUBLE_REAL
					glColor3dv(color[m].temp());
#else
					glColor3fv(color[m].temp());
#endif

#if FE_DOUBLE_REAL
				glNormal3dv(normal[m].temp());
				glVertex3dv(vertex[m].temp());
#else
				glNormal3fv(normal[m].temp());
				glVertex3fv(vertex[m].temp());
#endif
			}
		}
		else
		{
			if(color)
			{
				if(blending)
#if FE_DOUBLE_REAL
					glColor4dv(color[0].temp());
#else
					glColor4fv(color[0].temp());
#endif
				else
#if FE_DOUBLE_REAL
					glColor3dv(color[0].temp());
#else
					glColor3fv(color[0].temp());
#endif
			}

			if(grain>2)
			{
				for(U32 m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glNormal3dv(normal[m].temp());
					glVertex3dv(vertex[m].temp());
#else
					glNormal3fv(normal[m].temp());
					glVertex3fv(vertex[m].temp());
#endif
				}
			}
			else
			{
				for(U32 m=0; m < vertices; m++)
				{
#if FE_DOUBLE_REAL
					glVertex3d(vertex[m].temp()[0],vertex[m+1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3d(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#else
					glVertex3f(vertex[m].temp()[0],vertex[m+1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3f(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#endif

					m++;
#if FE_DOUBLE_REAL
					glVertex3d(vertex[m].temp()[0],vertex[m-1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3d(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#else
					glVertex3f(vertex[m].temp()[0],vertex[m-1].temp()[1],
							vertex[m].temp()[2]);
					glVertex3f(vertex[m].temp()[0],vertex[m].temp()[1],
							vertex[m].temp()[2]);
#endif
				}
			}
		}

		glEnd();
#endif

		if(pass)
		{
			setup(drawStyle(),lit,twoSidedLighting(),FALSE,
					frontfaceCulling(),backfaceCulling(),zBuffering());
			glDisable(GL_POLYGON_OFFSET_LINE);

#if DGL_LINEWIDTH
			if(pushLineWidth)
			{
				//* restore line width
				glPopAttrib();
			}
#endif
		}
	}

	if(blending)
	{
		glPopAttrib();
	}

#if DGL_CHECK_ERROR
	checkForError("DrawOpenGL::drawPolys complete");
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawAlignedText(const SpatialVector &location,
		const String text,const Color &color)
{
	const BWORD lit=isLit();

	if(!m_fontbase)
	{
		assureFont();
	}

	if(!m_fontbase)
	{
		feLog("DrawOpenGL::drawAlignedText fontbase==0\n");
		return;
	}

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

	if(lit)
	{
		glDisable(GL_LIGHTING);
	}

	const Real alpha=color[3];
	const BWORD blending=(alpha<1.0);
	if(blending)
	{
		pushBlending();
	}

#if FE_DOUBLE_REAL
	glColor4dv(color.temp());
	glRasterPos3dv(location.temp());
#else
	glColor4fv(color.temp());
	glRasterPos3fv(location.temp());
#endif

	glPushAttrib(GL_LIST_BIT);

	glListBase(m_fontbase);
	glCallLists(text.length(),GL_UNSIGNED_BYTE,(unsigned char *)text.c_str());

	glPopAttrib();

	if(blending)
	{
		glPopAttrib();
	}

	if(lit)
	{
		glEnable(GL_LIGHTING);
	}

//	glFinish();

#if DGL_CHECK_ERROR
	checkForError("DrawOpenGL::drawAlignedText complete");
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawSphereGLU(const SpatialTransform &transform,
		const SpatialVector *scale,const Color &color)
{
	const GLint slices=8;	// 6
	const GLint stacks=6;	// 4

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

#if DGL_ANTIALIAS_LINE
	BWORD antialiasAll=antialias() && drawStyle()<DrawMode::e_solid;
	if(antialiasAll)
		pushAntialiasLines();
#endif

#if DGL_LINEWIDTH
	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	if(pushLineWidth)
	{
		glPushAttrib(GL_LINE_BIT);
		glLineWidth(lineWidth());
	}
#endif

#if DGL_CHECK_ALPHA
	if(color[3]<=0.0f)
	{
		feX("DrawOpenGL::drawSphereGLU",
				"alpha<=0 is pointless\n  check your color: %s\n",
				c_print(color));
	}
#endif

	const Real alpha=color[3];
	const BWORD blending=(alpha<1.0);
	if(blending)
	{
		pushBlending();
	}
	glPushMatrix();
	Real raw[16];
	transform.copy16(raw);
#if FE_DOUBLE_REAL
	glMultMatrixd(raw);
#else
	glMultMatrixf(raw);
#endif

#if FE_DOUBLE_REAL
	glColor4dv(color.temp());
#else
	glColor4fv(color.temp());
#endif

	//* TODO scale appropriately

	gluSphere(m_pQuadric,(*scale)[0],slices,stacks);
	if(drawStyle()==DrawMode::e_edgedSolid)
	{
#if DGL_ANTIALIAS_LINE
		if(antialias())
			pushAntialiasLines();
#endif

		setup(DrawMode::e_outline,FALSE,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),zBuffering());
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(-1.0f, 0.0f);

		gluSphere(m_pQuadric,(*scale)[0],slices,stacks);

		setup(drawStyle(),TRUE,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),zBuffering());
		glDisable(GL_POLYGON_OFFSET_LINE);

#if DGL_ANTIALIAS_LINE
		if(antialias())
			glPopAttrib();
#endif
	}

	glPopMatrix();

	if(blending)
	{
		glPopAttrib();
	}

#if DGL_LINEWIDTH
	if(pushLineWidth)
	{
		//* restore line width
		glPopAttrib();
	}
#endif

#if DGL_ANTIALIAS_LINE
	if(antialiasAll)
		glPopAttrib();
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawCylinderGLU(const SpatialTransform &transform,
		const SpatialVector *scale,Real baseScale,const Color &color,
		U32 slices)
{
	const U32 stacks=1;

	if(slices<1)
	{
		slices=6;	// tweak
	}
	else if(slices<3)
	{
		slices=3;
	}

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

#if DGL_ANTIALIAS_LINE
	BWORD antialiasAll=antialias() && drawStyle()<DrawMode::e_solid;
	if(antialiasAll)
		pushAntialiasLines();
#endif
#if DGL_LINEWIDTH
	const BWORD pushLineWidth=(lineWidth()!=1.0f);
	if(pushLineWidth)
	{
		glPushAttrib(GL_LINE_BIT);
		glLineWidth(lineWidth());
	}
#endif

#if DGL_CHECK_ALPHA
	if(color[3]<=0.0f)
	{
		feX("DrawOpenGL::drawCylinderGLU",
				"alpha<=0 is pointless\n  check your color: %s\n",
				c_print(color));
	}
#endif

	const Real alpha=color[3];
	const BWORD lit=isLit();
	const BWORD blending=(alpha!=1.0f && lit);
	if(blending)
	{
		pushBlending();
	}

	glPushMatrix();
	Real raw[16];
	transform.copy16(raw);
#if FE_DOUBLE_REAL
	glMultMatrixd(raw);
#else
	glMultMatrixf(raw);
#endif

	if(blending)
	{
#if FE_DOUBLE_REAL
		glColor4dv(color.temp());
#else
		glColor4fv(color.temp());
#endif
	}
	else
	{
#if FE_DOUBLE_REAL
		glColor3dv(color.temp());
#else
		glColor3fv(color.temp());
#endif
	}

	//* TODO scale appropriately

	gluCylinder(m_pQuadric,
			(*scale)[0]*baseScale,	//* base radius
			(*scale)[0],			//* top radius
			(*scale)[2],			//* height
			slices,					//* around
			stacks);				//* along

	if(drawStyle()==DrawMode::e_edgedSolid)
	{
#if DGL_ANTIALIAS_LINE
		if(antialias())
			pushAntialiasLines();
#endif

		setup(DrawMode::e_outline,FALSE,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),zBuffering());
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(-1.0f, 0.0f);

		gluCylinder(m_pQuadric,
				(*scale)[0]*baseScale,	//* base radius
				(*scale)[0],			//* top radius
				(*scale)[2],			//* height
				slices,					//* around
				stacks);				//* along

		setup(drawStyle(),TRUE,twoSidedLighting(),FALSE,
				frontfaceCulling(),backfaceCulling(),zBuffering());
		glDisable(GL_POLYGON_OFFSET_LINE);

#if DGL_ANTIALIAS_LINE
		if(antialias())
			glPopAttrib();
#endif
	}

	glPopMatrix();

	if(blending)
	{
		glPopAttrib();
	}

#if DGL_LINEWIDTH
	if(pushLineWidth)
	{
		//* restore line width
		glPopAttrib();
	}
#endif

#if DGL_ANTIALIAS_LINE
	if(antialiasAll)
		glPopAttrib();
#endif

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::drawTransformedTriangles(const SpatialTransform &transform,
	const SpatialVector *scale,const SpatialVector *vertex,
	const SpatialVector *normal,const Vector2f *texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color *color,
	const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
	sp<DrawBufferI> spDrawBuffer)
{
	Real matrix[16];
	transform.copy16(matrix);

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

	glPushMatrix();
#if FE_DOUBLE_REAL
	glMultMatrixd(matrix);
#else
	glMultMatrixf(matrix);
#endif

	if(scale)
	{
		glScalef((*scale)[0],(*scale)[1],(*scale)[2]);
	}

	drawTriangles(vertex,normal,texture,vertices,strip,multicolor,
			color,hullPointMap,hullFacePoint,spDrawBuffer);
	glPopMatrix();

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::setupLights(void)
{
	m_brightness=fe::minimum(Real(1),fe::maximum(Real(0),m_brightness));
	m_contrast=fe::minimum(Real(2),fe::maximum(Real(0),m_contrast));

	const GLfloat position[]={ 1.0f, -1.0f, 3.0f, 0.0f };
	const GLfloat ambient[]=
			{ m_brightness, m_brightness, m_brightness, 1.0f };
	const GLfloat diffuse[]={ m_contrast, m_contrast, m_contrast, 1.0f };
	const GLfloat specular[]={ 1.0f, 1.0f, 1.0f, 1.0f };

	//* NOTE position light in camera space
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glLightfv(GL_LIGHT0,GL_POSITION,position);
	glLightfv(GL_LIGHT0,GL_AMBIENT,ambient);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuse);
	glLightfv(GL_LIGHT0,GL_SPECULAR,specular);

	glPopMatrix();

#if FALSE
	const SpatialVector direction=
			unitSafe(SpatialVector(position[0],position[1],position[2]));
	const GLfloat spot_direction[]=
			{ direction[0], direction[1], direction[2] };
	const GLfloat spot_exponent[]= { 0.0f };
	const GLfloat spot_cutoff[]= { 180.0f };

	glLightfv(GL_LIGHT0,GL_SPOT_DIRECTION,spot_direction);
	glLightfv(GL_LIGHT0,GL_SPOT_EXPONENT,spot_exponent);
	glLightfv(GL_LIGHT0,GL_SPOT_CUTOFF,spot_cutoff);
#endif

	// glScalef will affect lighting unless auto-normalized
	glEnable(GL_NORMALIZE);
}

void DrawOpenGL::setup(DrawMode::DrawStyle style,BWORD lit,BWORD dual,
		BWORD texture,BWORD frontcull,BWORD backcull,BWORD zBuffer)
{
	gluQuadricDrawStyle(m_pQuadric,
			drawStyle()==DrawMode::e_wireframe? GLU_SILHOUETTE:
			(drawStyle()==DrawMode::e_pointCloud? GLU_POINT: GLU_FILL));
	gluQuadricTexture(m_pQuadric,texture? GL_TRUE: GL_FALSE);

	if(style==DrawMode::e_outline)
	{
		glPolygonMode(GL_FRONT,GL_LINE);
		glPolygonMode(GL_BACK,GL_LINE);
	}
	else
	{
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	}

	if(lit)
	{
		if(texture)
		{
			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
//			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
//			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);

//			glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
			glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

			glEnable(GL_TEXTURE_2D);
//			glShadeModel(GL_FLAT);
		}
		else
		{
			glDisable(GL_TEXTURE_2D);
		}

#if FALSE
		GLfloat colormap[]={ 11.0f, 159.0f, 159.0f };
		glMaterialfv(GL_FRONT,GL_COLOR_INDEXES,colormap);
#endif

//		const GLfloat alpha=0.7f;

		if(drawStyle()==DrawMode::e_ghost)
		{
			//* TODO param
			const GLfloat specular[]={ 0.5f, 0.5f, 0.5f, 1.0f };
			const GLfloat shininess[]={ 32.0f };

			glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,specular);
			glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shininess);
		}
		else
		{
			//* default (off)
			const GLfloat specular[]={ 0.0f, 0.0f, 0.0f, 1.0f };
			glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,specular);
		}

#if FALSE
		const GLfloat black[]={ 0.0f, 0.0f, 0.0f, 1.0f };
		const GLfloat diffuse0[]={ 0.4f, 0.4f, 0.4f, alpha };

		const GLfloat ambient1[]={ 0.0f, 0.4f, 0.0f, alpha };
		const GLfloat diffuse1[]={ 0.0f, 0.6f, 0.0f, alpha };

		const GLfloat ambient2[]={ 0.0f, 0.0f, 0.4f, alpha };
		const GLfloat diffuse2[]={ 0.0f, 0.0f, 0.6f, alpha };

		if(dual)
		{
			glMaterialfv(GL_BACK,GL_AMBIENT,ambient2);
			glMaterialfv(GL_BACK,GL_DIFFUSE,diffuse2);
		}
		else
		{
			glMaterialfv(GL_BACK,GL_AMBIENT,ambient1);
			glMaterialfv(GL_BACK,GL_DIFFUSE,diffuse1);
		}

		if(style==DrawMode::e_edgedSolid)
		{
			glMaterialfv(GL_FRONT,GL_AMBIENT,black);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse0);
		}
		else
		{
			glMaterialfv(GL_FRONT,GL_AMBIENT,ambient1);
			glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse1);
		}
		glDisable(GL_COLOR_MATERIAL);
#else
		//* follow glColor calls
		glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
#endif

		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
		glShadeModel(GL_SMOOTH);
	}
	else
	{
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
	}

	//* Culling (Backface and/or Frontface)
//	if(frontcull || backcull || style==DrawMode::e_outline)
	if((frontcull || backcull) && style!=DrawMode::e_outline)
	{
		if(style==DrawMode::e_outline || (frontcull && backcull))
		{
			glCullFace(GL_FRONT_AND_BACK);
		}
		else if(backcull)
		{
			glCullFace(GL_BACK);
		}
		else
		{
			glCullFace(GL_FRONT);
		}
		glEnable(GL_CULL_FACE);
	}
	else
	{
		glDisable(GL_CULL_FACE);
	}

	//* Two Sided Lighting
	if(frontcull || backcull || style==DrawMode::e_outline || !dual)
	{
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
	}
	else
	{
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
	}

	if(zBuffer)
	{
		glEnable(GL_DEPTH_TEST);
	}
	else
	{
		glDisable(GL_DEPTH_TEST);
	}
}

void DrawOpenGL::createDefaultTexture(void)
{
	const long bit=0x8;
	unsigned char c;
	long m,n;
	for(m=0;m<64;m++)
	{
		for(n=0;n<64;n++)
		{
			c=128+127*(((m&bit)==0)^(((n&bit))==0));
			m_texture[m][n][0]=383-c;
			m_texture[m][n][1]=c;
			m_texture[m][n][2]=c;
		}
	}

	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
}

void DrawOpenGL::useDefaultTexture(void)
{
	glTexImage2D(GL_TEXTURE_2D,0,3,
			64,64,0,GL_RGB,GL_UNSIGNED_BYTE,
			m_texture);
}

void DrawOpenGL::queryFormat(sp<ImageI> spImageI,GLenum& format,
		GLenum& type,GLint& components)
{
	switch(spImageI->format())
	{
		case ImageI::e_colorindex:
			format=GL_COLOR_INDEX;
			type=GL_UNSIGNED_SHORT;
			components=1;
			break;
		case ImageI::e_rgba:
			format=GL_RGBA;
			type=GL_UNSIGNED_BYTE;
			components=4;
			break;
		default:
			format=GL_RGB;
			type=GL_UNSIGNED_BYTE;
			components=3;
			break;
	}
}

void DrawOpenGL::drawRaster(sp<ImageI> spImageI,const SpatialVector& location)
{
	if(spImageI.isNull())
	{
		return;
	}

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadLock();
	}
#endif

//	glPushAttrib(GL_DEPTH_BUFFER_BIT);
//	glDisable(GL_DEPTH_TEST);

	glRasterPos3f((GLfloat)location[0],(GLfloat)location[1],
			(GLfloat)location[2]);
	glPixelZoom(1.0f,-1.0f);

	GLenum format;
	GLenum type;
	GLint components;
	queryFormat(spImageI,format,type,components);

	glDrawPixels((GLsizei)spImageI->width(),(GLsizei)spImageI->height(),
			format,type,(const GLvoid*)spImageI->raw());

//	glPopAttrib();

#if DGL_THREAD_LOCKING
	if(m_lockThreads)
	{
		threadUnlock();
	}
#endif
}

void DrawOpenGL::bindTexture(sp<ImageI> spImageI)
{
	if(spImageI==m_currentTexture)
	{
		return;
	}

	m_currentTexture=spImageI;

	if(!spImageI.isValid())
	{
		useDefaultTexture();
		return;
	}

	GLenum format;
	GLenum type;
	GLint components;
	queryFormat(spImageI,format,type,components);

	const GLenum target=GL_TEXTURE_2D;
	const GLint levels=0;
	const GLint border=0;
	glTexImage2D(target,levels,components,
			(GLsizei)spImageI->width(),(GLsizei)spImageI->height(),border,
			format,type,(const GLvoid*)spImageI->raw());
}

} // namespace
