/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SAR_DEBUG			FALSE
#define FE_SAR_CLONE_DEBUG		FALSE
#define FE_SAR_SPEC_DEBUG		FALSE

#include "surface.pmh"

namespace fe
{

SurfaceAccessibleRecord::SurfaceAccessibleRecord(void)
{
}

SurfaceAccessibleRecord::~SurfaceAccessibleRecord(void)
{
}

void SurfaceAccessibleRecord::reset(void)
{
	if(m_spScope.isValid())
	{
		SurfaceModel surfaceModelRV;
		surfaceModelRV.bind(m_spScope);
		m_weakRecord=surfaceModelRV.createRecord();

		FEASSERT(m_weakRecord.isValid());

		m_spRecordGroup=new RecordGroup();
		m_spRecordGroup->add(m_weakRecord);
	}
	else
	{
		//* TODO check this
		FEASSERT(FALSE);

		m_weakRecord=Record();
		m_spRecordGroup=NULL;
	}

	m_spForeignAccessible=NULL;
}

void SurfaceAccessibleRecord::bind(Record& a_rRecord)
{
	if(!a_rRecord.isValid())
	{
		reset();
		return;
	}

	m_spScope=a_rRecord.layout()->scope();

	//* check if just a payload surface (not SurfaceAccessibleRecord)
	sp<Layout> spSurfaceLayout=m_spScope->declare("surface");
	m_spScope->support("accessible","Component");
	m_spScope->populate("surface","accessible");

	Accessor< sp<Component> > accessorAccessible;
	accessorAccessible.setup(m_spScope,"accessible");

	if(accessorAccessible.check(a_rRecord))
	{
		reset();

#if TRUE
		m_spForeignAccessible=accessorAccessible(a_rRecord);
#else
		copy(accessorAccessible(a_rRecord));
		save("foreign.rg");
#endif

		feLog("SurfaceAccessibleRecord::bind foreign surface %p \"%s\"\n",
				m_spForeignAccessible.raw(),
				m_spForeignAccessible->name().c_str());
	}
	else
	{
		m_weakRecord=a_rRecord;

		m_spRecordGroup=new RecordGroup();
		m_spRecordGroup->add(m_weakRecord);

		m_spForeignAccessible=NULL;
	}
}

Protectable* SurfaceAccessibleRecord::clone(Protectable* pInstance)
{
#if	FE_SAR_CLONE_DEBUG
	feLog("SurfaceAccessibleRecord::clone\n");

	const U32 us1=systemMicroseconds();
#endif

	FEASSERT(m_spScope.isValid());

	SurfaceAccessibleRecord* pSurfaceAccessibleRecord= pInstance?
			dynamic_cast<SurfaceAccessibleRecord*>(pInstance):
			new SurfaceAccessibleRecord();

	pSurfaceAccessibleRecord->setLibrary(library());

	//* NOTE would be risky to share scope

	sp<Scope> spScope=registry()->create("Scope");
	pSurfaceAccessibleRecord->bind(spScope);

	sp<RecordGroup>& rspRecordGroup=pSurfaceAccessibleRecord->m_spRecordGroup;

#if	FE_SAR_CLONE_DEBUG
	const U32 us2=systemMicroseconds();
#endif

	//* why is direct cloning slower?
//	Cloner cloner(m_spScope);
//	rspRecordGroup=cloner.clone(m_spRecordGroup);

	sp<data::StreamI> spStream=
			sp<data::StreamI>(new data::BinaryStream(m_spScope));

	std::stringstream stringStream;
	spStream->output(stringStream,m_spRecordGroup);

#if	FE_SAR_CLONE_DEBUG
	const U32 us3=systemMicroseconds();
#endif

	rspRecordGroup=spStream->input(stringStream);

#if	FE_SAR_CLONE_DEBUG
	const U32 us4=systemMicroseconds();
#endif

	//* update surface with new Record
	RecordGroup::iterator it=rspRecordGroup->begin();
	sp<RecordArray> spRA= *it;
	Record record=spRA->getRecord(0);
	pSurfaceAccessibleRecord->bind(record);

	FEASSERT(pSurfaceAccessibleRecord->m_spScope.isValid());

#if	FE_SAR_CLONE_DEBUG
	const U32 us5=systemMicroseconds();

	feLog("SurfaceAccessibleRecord::clone us %d %d %d %d -> %d\n",
			us2-us1,us3-us2,us4-us3,us5-us4,us5-us1);
#endif

	return pSurfaceAccessibleRecord;
}

BWORD SurfaceAccessibleRecord::load(String a_filename,sp<Catalog> a_spSettings)
{
#if	FE_SAR_DEBUG
	feLog("SurfaceAccessibleRecord::load \"%s\"\n",a_filename.c_str());
#endif

	BWORD binary=FALSE;

	if(a_spSettings.isValid())
	{
		const String options=a_spSettings->catalog<String>("options");
		binary=options.match(".*\\bbinary\\b.*");
	}

	FEASSERT(m_spScope.isValid());

//	if(a_filename.match(".*\\.obj"))
	if(a_filename.chop(".obj")!=a_filename)
	{
		String buffer;
		buffer+="DEFAULTGROUP 1\n";
		buffer+="RECORD Surface0 SurfaceOBJ\n";
		buffer+="	surf:filename \""+a_filename+"\"\n";
		buffer+="END\n";

		m_spRecordGroup=
				RecordView::loadRecordGroupFromBuffer(m_spScope,buffer,FALSE);
	}
	else
	{
		m_spRecordGroup=
				RecordView::loadRecordGroup(m_spScope,a_filename,binary);
	}

	if(m_spRecordGroup.isNull())
	{
		feLog("SurfaceAccessibleRecord::load NULL RecordGroup\n");
		return FALSE;
	}

	RecordGroup::iterator it=m_spRecordGroup->begin();
	sp<RecordArray> spRA= *it;

	FEASSERT(spRA.isValid());
	m_weakRecord=spRA->getRecord(0);

	outlineCreateDefault();

	return TRUE;
}

BWORD SurfaceAccessibleRecord::save(String a_filename,sp<Catalog> a_spSettings)
{
#if	FE_SAR_DEBUG
	feLog("SurfaceAccessibleRecord::save\n");
#endif

	FEASSERT(m_spScope.isValid());
	FEASSERT(m_spRecordGroup.isValid());

	if(m_spScope.isNull() || m_spRecordGroup.isNull())
	{
		return FALSE;
	}

	const String options=a_spSettings.isValid()?
			a_spSettings->catalog<String>("options"): "";
	const BWORD binary=options.match(".*\\bbinary\\b.*");

	RecordView::saveRecordGroup(m_spScope,m_spRecordGroup,a_filename,binary);
	return TRUE;
}

void SurfaceAccessibleRecord::groupNames(Array<String>& a_rNameArray,
	SurfaceAccessibleI::Element a_element) const
{
	a_rNameArray.clear();

	//* TODO
}

void SurfaceAccessibleRecord::attributeSpecs(
	Array<SurfaceAccessibleI::Spec>& a_rSpecs,
	SurfaceAccessibleI::Element a_element) const
{
	if(m_spForeignAccessible.isValid())
	{
		return m_spForeignAccessible->attributeSpecs(a_rSpecs,a_element);
	}

	const String layoutName=elementLayout(a_element);

#if FE_SAR_SPEC_DEBUG
	feLog("SurfaceAccessibleRecord::attributeSpecs layout %s\n",
			layoutName.c_str());
#endif

	a_rSpecs.clear();

	SurfacePoint surfaceElementRV;

	if(a_element==SurfaceAccessibleI::e_point)
	{
		SurfaceModel surfaceModelRV;
		surfaceModelRV.bind(m_weakRecord);

		sp<RecordGroup> spRG;
		switch(a_element)
		{
			case SurfaceAccessibleI::e_point:
				spRG=surfaceModelRV.surfacePointRG();
				break;
			case SurfaceAccessibleI::e_vertex:
				spRG=surfaceModelRV.surfaceVertexRG();
				break;
			case SurfaceAccessibleI::e_primitive:
				spRG=surfaceModelRV.surfacePrimitiveRG();
				break;
			case SurfaceAccessibleI::e_detail:
				spRG=surfaceModelRV.surfaceDetailRG();
				break;
			default:
				return;
		}

		if(spRG.isNull() || !spRG->count())
		{
#if FE_SAR_SPEC_DEBUG
			feLog("SurfaceAccessibleRecord::attributeSpecs layout %s no data\n",
					layoutName.c_str());
#endif
			return;
		}

		RecordGroup::iterator it=spRG->begin();
		sp<RecordArray> spRecordArray= *it;
		if(spRecordArray.isNull())
		{
			return;
		}

		surfaceElementRV.bind(spRecordArray);
	}

	sp<Layout> spElementLayout=m_spScope->lookupLayout(layoutName);
	sp<TypeMaster> spTypeMaster=registry()->master()->typeMaster();

#if FE_SAR_SPEC_DEBUG
	feLog("SurfaceAccessibleRecord::attributeSpecs layout %s %p\n",
			layoutName.c_str(),spElementLayout.raw());
#endif

	if(spElementLayout.isValid())
	{
		const U32 attrCount=spElementLayout->attributeCount();

#if FE_SAR_SPEC_DEBUG
		feLog("SurfaceAccessibleRecord::attributeSpecs attrCount %d\n",
				attrCount);
#endif

		for(U32 attrIndex=0;attrIndex<attrCount;attrIndex++)
		{
			sp<fe::Attribute> spAttribute=spElementLayout->attribute(attrIndex);
			const String attrName=spAttribute->name();

			if(attrName==":CT" || attrName==":SN")
			{
				continue;
			}

			if(a_element==SurfaceAccessibleI::e_point)
			{
				//* TODO check other attributes
				if(attrName=="P" &&
						!surfaceElementRV.at.check())
				{
					//* NOTE probably merits an exception
					continue;
				}
				if(attrName=="N" &&
						!surfaceElementRV.up.check())
				{
					continue;
				}
				if(attrName=="uvw" &&
						!surfaceElementRV.uvw.check())
				{
					continue;
				}
				if(attrName=="Cd" &&
						!surfaceElementRV.color.check())
				{
					continue;
				}
			}

			const String attrTypeName=spTypeMaster->lookupBaseType(
					spAttribute->type());
			String typeName=attrTypeName;	//* string, color

			if(attrTypeName=="I32")
			{
				typeName="integer";
			}
			else if(attrTypeName=="F32")
			{
				typeName="real";
			}
			else if(attrTypeName=="vector3f")
			{
				typeName="vector3";
			}

#if FE_SAR_SPEC_DEBUG
			feLog("SurfaceAccessibleRecord::attributeSpecs %d/%d %s %s %s\n",
					attrIndex,attrCount,
					attrTypeName.c_str(),typeName.c_str(),attrName.c_str());
#endif

			SurfaceAccessibleI::Spec spec;
			spec.set(commonName(attrName),typeName);

			a_rSpecs.push_back(spec);
		}
	}
}

void SurfaceAccessibleRecord::require(SurfaceAccessibleI::Element a_element,
	String a_attrName,String a_attrTypeName)
{
	const String layoutName=elementLayout(a_element);

	m_spScope->declare(layoutName);
	m_spScope->support(a_attrName,a_attrTypeName);
	m_spScope->populate(layoutName,a_attrName);
}

sp<SurfaceAccessorI> SurfaceAccessibleRecord::accessor(
	String a_node,Element a_element,String a_name,Creation a_create)
{
	if(m_spForeignAccessible.isValid())
	{
		return m_spForeignAccessible->accessor(a_node,a_element,
				a_name,a_create);
	}

	FEASSERT(m_weakRecord.isValid());

	sp<SurfaceAccessorRecord> spAccessor(new SurfaceAccessorRecord);

	if(spAccessor.isValid())
	{
		FEASSERT(m_weakRecord.isValid());

		spAccessor->bind(m_weakRecord);
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));

		if(spAccessor->bind(a_element,a_name) ||
				a_create==SurfaceAccessibleI::e_createMissing)
		{
			return spAccessor;
		}
	}

	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleRecord::accessor(
	String a_node,Element a_element,Attribute a_attribute,Creation a_create)
{
	if(m_spForeignAccessible.isValid())
	{
		return m_spForeignAccessible->accessor(a_node,a_element,
				a_attribute,a_create);
	}

	if(!m_weakRecord.isValid())
	{
		feX(e_notInitialized,"no data");
	}

	FEASSERT(m_weakRecord.isValid());

	sp<SurfaceAccessorRecord> spAccessor(new SurfaceAccessorRecord);

	if(spAccessor.isValid())
	{
		FEASSERT(m_weakRecord.isValid());

		spAccessor->bind(m_weakRecord);

		FEASSERT(m_weakRecord.isValid());

		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));

		FEASSERT(m_weakRecord.isValid());

		if(spAccessor->bind(a_element,a_attribute) ||
				a_create==SurfaceAccessibleI::e_createMissing)
		{
			FEASSERT(m_weakRecord.isValid());

			return spAccessor;
		}

//		feLog("SurfaceAccessibleRecord::accessor"
//				" failed element %d attr %d create %d\n",
//				a_element,a_attribute,a_create);
	}

	FEASSERT(m_weakRecord.isValid());

	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceI> SurfaceAccessibleRecord::surface(String a_group)
{
#if	FE_SAR_DEBUG
	feLog("SurfaceAccessibleRecord::surface(\"%s\")\n",a_group.c_str());
#endif

	if(m_spForeignAccessible.isValid())
	{
		feLog("SurfaceAccessibleRecord::surface foreign delegation\n");

		return m_spForeignAccessible->surface(a_group);
	}

	//* NOTE some RecordGroup files create their own SurfaceI component
	if(m_weakRecord.isValid())
	{
		SurfaceModel surfaceModelRV;
		surfaceModelRV.bind(m_weakRecord);

		cp<SurfaceI> cpSurfaceI=surfaceModelRV.component();

		if(cpSurfaceI.isValid())
		{
			return cpSurfaceI;
		}
	}

	return SurfaceAccessibleBase::surface(a_group);
}

sp<SurfaceI> SurfaceAccessibleRecord::subSurface(U32 a_subIndex,String a_group)
{
	//* TODO

	return sp<SurfaceI>(NULL);
}

void SurfaceAccessibleRecord::copy(
	sp<SurfaceAccessibleI> a_spSurfaceAccessible)
{
	sp<SurfaceAccessibleRecord> spSurfaceAccessibleRecord(
			a_spSurfaceAccessible);
	if(spSurfaceAccessibleRecord.isValid() &&
			spSurfaceAccessibleRecord->m_spForeignAccessible.isValid())
	{
		reset();

		//* NOTE not really a deep copy
		m_spForeignAccessible=spSurfaceAccessibleRecord->m_spForeignAccessible;
		return;
	}

	reset();
	append(a_spSurfaceAccessible,NULL);
}

void SurfaceAccessibleRecord::copy(
	sp<SurfaceAccessibleI> a_spSurfaceAccessible,sp<FilterI> a_spFilter)
{
	sp<SurfaceAccessibleRecord> spSurfaceAccessibleRecord(
			a_spSurfaceAccessible);
	if(spSurfaceAccessibleRecord.isValid() &&
			spSurfaceAccessibleRecord->m_spForeignAccessible.isValid())
	{
		reset();

		//* NOTE not really a deep copy
		m_spForeignAccessible=spSurfaceAccessibleRecord->m_spForeignAccessible;
		return;
	}

	reset();
	append(a_spSurfaceAccessible,NULL,a_spFilter);
}

} // namespace
