//
// Copyright 2016 Pixar
//
// Licensed under the Apache License, Version 2.0 (the "Apache License")
// with the following modification; you may not use this file except in
// compliance with the Apache License and the following modification to it:
// Section 6. Trademarks. is deleted and replaced with:
//
// 6. Trademarks. This License does not grant permission to use the trade
//	names, trademarks, service marks, or product names of the Licensor
//	and its affiliates, except as required to comply with Section 4(c) of
//	the License and to reproduce the content of the NOTICE file.
//
// You may obtain a copy of the Apache License at
//
//	 http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the Apache License with the above modification is
// distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied. See the Apache License for the specific
// language governing permissions and limitations under the Apache License.

// originally from pxr/imaging/lib/hd/unitTestHelper.cpp
// adapted for Free Electron

#include "usd/hydra.pmh"

#include "pxr/imaging/hd/rprimCollection.h"
#include "pxr/imaging/hd/tokens.h"

#include "pxr/imaging/glf/diagnostic.h"

#include "pxr/base/gf/matrix4d.h"
#include "pxr/base/gf/frustum.h"
#include "pxr/base/tf/getenv.h"
#include "pxr/base/tf/staticTokens.h"

#include <boost/functional/hash.hpp>

#include <string>
#include <sstream>

HydraDriver::HydraDriver(void):
	m_delegate(HdRenderIndex::New(&m_renderDelegate),
	SdfPath::AbsoluteRootPath()),
	m_renderPassState(new HdRenderPassState())
{
	TfToken reprName = HdTokens->hull;
	if(TfGetenv("HD_ENABLE_SMOOTH_NORMALS", "CPU") == "CPU" or
			TfGetenv("HD_ENABLE_SMOOTH_NORMALS", "CPU") == "GPU")
	{
		reprName = HdTokens->smoothHull;
	}
	_Init(reprName);
}

HydraDriver::HydraDriver(TfToken const &reprName):
	m_delegate(HdRenderIndex::New(&m_renderDelegate),
	SdfPath::AbsoluteRootPath()),
	m_renderPassState(new HdRenderPassState())
{
	_Init(reprName);
}

void HydraDriver::_Init(TfToken const &reprName)
{
	m_reprName = reprName;

	GfMatrix4d viewMatrix = GfMatrix4d().SetIdentity();
	viewMatrix *= GfMatrix4d().SetTranslate(GfVec3d(0.0, 1000.0, 0.0));
	viewMatrix *= GfMatrix4d().SetRotate(
			GfRotation(GfVec3d(1.0, 0.0, 0.0), -90.0));

	GfFrustum frustum;
	frustum.SetPerspective(45, true, 1, 1.0, 10000.0);
	GfMatrix4d projMatrix = frustum.ComputeProjectionMatrix();

	SetCamera(viewMatrix, projMatrix, GfVec4d(0, 0, 512, 512));

	// set depthfunc to GL default
	m_renderPassState->SetDepthFunc(HdCmpFuncLess);
}

void HydraDriver::Draw(HdEngine& a_rEngine,bool withGuides)
{
	Draw(a_rEngine,GetRenderPass(withGuides));
}

void HydraDriver::Draw(HdEngine& a_rEngine,
	HdRenderPassSharedPtr const &renderPass)
{
	//* protect OpenGL state
//	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glPushAttrib(GL_COLOR_BUFFER_BIT);

	VtValue vtValue(m_renderPassState);
	a_rEngine.SetTaskContextData(HdxTokens->renderPassState,vtValue);

	a_rEngine.Execute(m_delegate.GetRenderIndex(),
			m_delegate.GetRenderTasks());

	glPopAttrib();

	GLF_POST_PENDING_GL_ERRORS();
}

void HydraDriver::SetCamera(GfMatrix4d const &modelViewMatrix,
	GfMatrix4d const &projectionMatrix,GfVec4d const &viewport)
{
	m_delegate.SetCamera(modelViewMatrix,projectionMatrix);

	m_renderPassState->SetCamera(modelViewMatrix,
			projectionMatrix,viewport);
}

HdRenderPassSharedPtr const& HydraDriver::GetRenderPass(bool withGuides)
{
	if(withGuides)
	{
		if(not m_geomAndGuidePass)
		{
			m_geomAndGuidePass = HdRenderPassSharedPtr(
					new NullRenderPass(&m_delegate.GetRenderIndex(),
					HdRprimCollection(TestTokens->geometryAndGuides,
					m_reprName)));
		}
		return m_geomAndGuidePass;
	}
	else
	{
		if(not m_geomPass)
		{
			m_geomPass = HdRenderPassSharedPtr(
					new NullRenderPass(&m_delegate.GetRenderIndex(),
					HdRprimCollection(HdTokens->geometry,m_reprName)));
		}
		return m_geomPass;
	}
}

void HydraDriver::SetRepr(TfToken const &reprName)
{
	m_reprName = reprName;

	if(m_geomAndGuidePass)
	{
		m_geomAndGuidePass->SetRprimCollection(
				HdRprimCollection(TestTokens->geometryAndGuides, m_reprName));
	}

	if(m_geomPass)
	{
		m_geomPass->SetRprimCollection(
				HdRprimCollection(HdTokens->geometry, m_reprName));
	}
}
