/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "fe/platform.h"

namespace fe
{

FE_DL_PUBLIC RegexInitFunction* Regex::ms_fnInit=NULL;
FE_DL_PUBLIC RegexMatchFunction* Regex::ms_fnMatch=NULL;
FE_DL_PUBLIC RegexMatchFunction* Regex::ms_fnSearch=NULL;
FE_DL_PUBLIC RegexResultFunction* Regex::ms_fnResult=NULL;
FE_DL_PUBLIC RegexReplaceFunction* Regex::ms_fnReplace=NULL;
FE_DL_PUBLIC RegexFinishFunction* Regex::ms_fnFinish=NULL;
FE_DL_PUBLIC RegexReleaseFunction* Regex::ms_fnRelease=NULL;
FE_DL_PUBLIC BWORD Regex::ms_checked=FALSE;

Regex::Regex(const char* a_pattern)
{
	confirm();

	m_pResult=NULL;

	if(ms_fnInit)
	{
		m_pExpression=ms_fnInit(a_pattern);
		return;
	}

	m_pExpression=strdup(a_pattern);
}

Regex::~Regex(void)
{
	if(m_pResult)
	{
		if(ms_fnRelease)
		{
			ms_fnRelease(m_pResult);
		}
		m_pResult=NULL;
	}
	if(m_pExpression)
	{
		if(ms_fnFinish)
		{
			ms_fnFinish(m_pExpression);
		}
		m_pExpression=NULL;
	}
}

BWORD Regex::match(const char* a_candidate)
{
	if(m_pResult)
	{
		FEASSERT(ms_fnRelease);
		ms_fnRelease(m_pResult);
		m_pResult=NULL;
	}

	if(ms_fnMatch)
	{
		m_pResult=ms_fnMatch(m_pExpression,a_candidate);
		return m_pResult!=NULL;
	}

	const char* pattern=(const char*)m_pExpression;

	if(strlen(pattern)==2 && !strcmp(pattern,".*"))
	{
		return TRUE;
	}

	if(!strcmp(pattern,a_candidate))
	{
		return TRUE;
	}

	//* TEXT.*
	const I32 len=strlen(pattern);
	if(!strncmp(&pattern[len-2],".*",2) &&
			!strncmp(pattern,a_candidate,len-2))
	{
		return TRUE;
	}

	//* TODO consider other common patterns

//	printf("Regex::match miss \"%s\" \"%s\"\n",pattern,a_candidate);

	return FALSE;
}

BWORD Regex::search(const char* a_candidate)
{
	if(m_pResult)
	{
		FEASSERT(ms_fnRelease);
		ms_fnRelease(m_pResult);
		m_pResult=NULL;
	}

	if(ms_fnSearch)
	{
		m_pResult=ms_fnSearch(m_pExpression,a_candidate);
		return m_pResult!=NULL;
	}

	return match(a_candidate);
}

const char* Regex::replace(const char* a_candidate,const char* a_replacement)
{
	if(m_pResult)
	{
		FEASSERT(ms_fnRelease);
		ms_fnRelease(m_pResult);
		m_pResult=NULL;
	}

	if(ms_fnReplace)
	{
		m_pResult=ms_fnReplace(m_pExpression,a_candidate,a_replacement);
		if(m_pResult)
		{
			return result(0);
		}
	}

	return a_candidate;
}

const char* Regex::result(U32 a_index)
{
	if(ms_fnSearch)
	{
		FEASSERT(ms_fnResult);
		return ms_fnResult(m_pResult,a_index);
	}

	return "";
}

//* static
BWORD Regex::confirm(void)
{
	if(ms_fnInit)
	{
		return TRUE;
	}
	if(ms_checked)
	{
		return FALSE;
	}
	ms_checked=TRUE;

	const String verbose=System::getVerbose();

	System::printBannerOnce();

	DL_Loader *pLoader=new DL_Loader();
	FEASSERT(pLoader);

	const BWORD adaptname=TRUE;

	String nameOrder=FE_STRING(FE_REGEX_ORDER);
	String name;
	while(!(name=nameOrder.parse("\"",":")).empty())
	{
		if(!pLoader->openDL(name,adaptname))
		{
			pLoader->closeDL();
			continue;
		}

		bool (*regex_init)(void);
		regex_init=(bool (*)(void))pLoader->findDLFunction("regex_init");

		if(!regex_init)
		{
			if(verbose!="none")
			{
				fprintf(stderr,"%sregex_init not found in \"%s\"\n",
						PrefixLog::prefix().c_str(),name.c_str());
			}

			pLoader->closeDL();
			continue;
		}

		if(regex_init())
		{
			if(verbose=="all")
			{
				printf("%susing %s (for regular expressions)\n",
						PrefixLog::prefix().c_str(),name.c_str());
			}

			return TRUE;

			//* TODO retain, close, and delete pLoader later
		}
	}

	if(verbose!="none")
	{
		printf("%susing simplified regex\n",PrefixLog::prefix().c_str());
	}

	delete pLoader;
	return FALSE;
}

} /* namespace */

